/* *** August 21, 2002: Convention for y-reflection changed
   *** September 6, 2002: Inline version of state_to_index() implemented
               compile - without defining INDEX_INLINE for testing/debugging
                       - with "-DINDEX_INLINE" for production codes
   *** September 17, 2002: Preparations for parallelization:
         - memory allocation for states replaced by configuration on stack
         - all matrices (Hamiltonian, Correlators) replaced by transpose
           (which is the same, since the matrices are symmetric)
         - scalar complex variables (result, resultpm) in matrix loops introduced 
   *** September 18, 2002: Parallelization pragmas added
   *** September 19, 2002:
         - state_to_index() macro shortened (1 variable eliminated)
         - possibility of fixing ysize at compile time implemented
   *** September 20/21, 2002: Some reorganization in apply_internal_ham()
         (thanks to Josef Schuele for hints)
   *** September 21, 2002: Speedup in non-parallel part of create_momenta()
         (x- and y-loops exchanged ...)
   *** September 24/25, 2002: (partially following hints of Josef Schuele)
         - matrix multiplication in apply_internal_ham() buffered linewise
         - data types of index_tab, index, ... changed to smaller ones on
           64-Bit machines
         - possibility of fixing xsize at compile time added
   *** May 13, 2003: Added (optional) change of XXZ-anisotropy Delta with
         a line "z=..." in the file "internal-params"
   *** March 16, 2004: data type "long" for computation of binomial coefficients
         replaced by "long long"
   *** January 1, 2004: data type LLONG introduced to allow handling systems
         with bigger linear sizes
*/

/* For a fixed x-size define xsize here
   (if not defined here, the x-size is defined during runtime using
    specifications in the input-file "internal-params")
*/

/* #define xsize 6 */

/* For a fixed y-size define ysize here
   (if not defined here, the y-size is defined during runtime using
    specifications in the input-file "internal-params")
*/

/* #define ysize 6 */

/* Size of buffer of line of matrix elements
   NOTE: The program can only work properly if this is at least as large
         as the maximal number of non-zero matrix elements in a row */

#define LBUFSIZ 2048

/* **** Includes **** */

/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "matrix.h"
#include "vectors.h"
#include "apply_h.h"

/* data type for spin representation -
   this defines the available linear extent */

/* typedef unsigned long LLONG; */
typedef unsigned long long LLONG;

/* The parameters */

double Delta = 1.0L;
double jval = 1.0L;
double alpha = 0.5L;

/* The maximal number of states we can handle */

/* in the Sz-subspace */

#define MAXNSTATES ( ((unsigned long) 1) << (sizeof(LONG)*8 - 1) ) + \
     ( ( ((unsigned long) 1) << (sizeof(LONG)*8 - 1) ) - 1)

/* ... and the momentum-subspaces */

#define MAXSNSTATES ( ((unsigned long) 1) << (sizeof(SLONG)*8 - 1) ) + \
     ( ( ((unsigned long) 1) << (sizeof(SLONG)*8 - 1) ) - 1000)


/* Two special numbers (should be negative and different, otherwise arbitrary) */

#define MAGIC (MAXSNSTATES+123)
#define MAGIX (MAXSNSTATES+876)

/* The maximal linear extent */

#define MAXD (8*sizeof(LLONG))

/* The maximal y-size */

#ifndef ysize
#define MAXY MAXD
#else
#define MAXY ysize
#endif

/* The size in x-direction and y-direction */

#ifndef xsize
static int xsize=-1;
#endif
#ifndef ysize
static int ysize=-1;
#endif

/* The total number of sites */

static int sites;

/* The number of particles (corresponding to total S^z) */

static int n_part=-1;

/* The various parities */

static int momentum_x=1;	/* x-momentum */
static int momentum_y=1;	/* y-momentum */
static int parity_x=1;		/* reflection x */
static int parity_y=1;		/* reflection y */
static int parity_d=0;		/* diagonal parity */

/* The number of states in the total-S^z vector space */

static unsigned long nstates;

/* Dimension of the momentum eigenspace */

static long pstates;

/* Mapping tables */

static SLONG *charge_momentum;
	/* Which S^z eigenvector belongs to which momentum eigenvector ? */

/* ********************************************************
   For a small system, you may wish to change the data type
   of charge_prefac to a shorter one, i.e.
      signed short   ->   signed char */

static signed short *charge_prefac;
	/* The (sign) prefactors */

static LONG *momentum_charge;
	/* Representative S^z eigenvector for the momentum eigenvector */

static INT *momentum_prefac;
	/* The number of states associated to the momentum eigenvector */

/* Mapping tables */

#ifndef SLOWINDEX
#ifdef xsize
#define tab_size (((LLONG 1)<<xsize)
LONG *index_tab[tab_size][ysize];
#else
LONG ***index_tab;
int tab_size;
#endif
char *np_tab;

#endif

/* Table of binomial coefficients */

static long *bin_coeff[4500];

/* ****** Some math routines ****** */

/* define DIRECTBIN in order not to store, but to compute the binomial
   coefficients - useful for very large systems at small densities  */

#ifdef DIRECTBIN
  #define over(a,b) binomial(a,b)
#else
  #define over(a,b) bin_coeff[a][b]
#endif

/* Den groessten gemeinsamen Teiler zweier Zahlen feststellen
   ueber Euklidischen Algorithmus
   Zehnerpotenzen schneller als Primfaktorzerlegung */

static long long get_gcd(long long a, long long b)
 {
   register int i;
   long long remainder;

   if(a == 0)           /* Sonderbehandlung fuer Nullen */
     return(1);
   if(b == 0)
     return(1);

   if(a < 0)            /* Betrag bilden */
     a=-a;
   if(b < 0)
     b=-b;

   if(a == 1)           /* Sonderbehandlung fuer Einsen */
     return(1);
   if(b == 1)
     return(1);

   remainder = a % b;   /* Das ist der Euklische Algorithmus */
   while(remainder)
    {
      a=b;
      b=remainder;
      remainder = a % b;
    }           /* Nun steht der GGT in b */

   return(b);
 }

/* Der Binomialkoeffizient */

static long binomial(long n, long m)
 {
   register long i;
   register long long gcd;
   register long long res1, res2;

   if(m > n-m)                  /* Symmetrie ausnutzen */
     m = n-m;

   if(m < 0)
     return(0);

#ifdef PRECOMP
   if(n==36)
    {
     if(m==10)
       return(254186856L);
     if(m==11)
       return(600805296L);
    }

   if(m==11)
    {
     if(n==34)
       return(286097760L);
     if(n==35)
       return(417225900L);
    }

   if(m==5)
    {
     if(n==141)
        return(432295143L);
     if(n==142)
        return(448072338L);
     if(n==143)
        return(464306843L);
     if(n==144)
        return(481008528L);
    }
#endif

   res1 = 1;
   res2 = 1;
   for(i=0; i<m; i++)
    {
      res1 *= (n-i);           /* Multiplizieren */
      res2 *= (i+1);
      gcd = get_gcd(res1, res2);      /* Und kuerzen */
      res1 /= gcd;
      res2 /= gcd;
    }
   if(res2 != 1)
     fprintf(stderr, "*** Error: Binomial coefficient (%ld,%ld) not an integer !\n", n, m);
   return(res1);
 }

/* Allocate memory and produce error message if not available */

static void *m_alloc(long size, char *str)
 {
  void *ptr;

  ptr = (void *)malloc(size);

  if(! ptr)
   {
    fprintf(stderr, "### Error: Out of memory in %s()\n", str);
    exit(-13);
   }
/*
  fprintf(stderr, "Allocated %ld bytes in routine %s()\n", size, str);
*/
  return(ptr);
 }

/* Initialize binomial coefficients */

static void init_binomials()
 {
  register long n, m;

  for(n=0; n<=sites; n++)
   {
    bin_coeff[n] = (long *)m_alloc((((n+1)<n_part)?(n+2):(n_part+1))*sizeof(long), "init_binomials");
    for(m=0; m<=n+1; m++)
     if(m <= n_part)
       bin_coeff[n][m] = binomial(n,m);
   }
 }

/* Free binomial coefficients */

static void free_binomials()
 {
  register long n;

  for(n=0; n<=sites; n++)
    free(bin_coeff[n]);
 }

#ifndef SLOWINDEX

static char IMSTR[] = "init_mtables";

/* Initialize mapping tables */

static void init_mtables()
 {
  register int i, j, x;
  LONG index;
  LLONG mask;
  int cnt, np, lnp;

#ifndef xsize
  tab_size = ((LLONG) 1)<<xsize;
#endif

  np_tab = (char *)m_alloc(tab_size * sizeof(char), IMSTR);

  for(i=0; i<tab_size; i++)
   {
    cnt = 0;
    mask = 1;
    for(x = 0; x<xsize; x++)
     {
       if(i & mask)
         cnt++;
       mask = mask<<1;
     }
    np = np_tab[i] = cnt;
   }

#ifndef xsize
  index_tab = (LONG ***)m_alloc(tab_size * sizeof(LONG **), IMSTR);
#endif

  for(i=0; i<tab_size; i++)
   {
#ifndef xsize
    index_tab[i] = (LONG **)m_alloc(ysize * sizeof(LONG *), IMSTR);
#endif
    for(j=0; j<ysize; j++)
     {
      if(n_part - np_tab[i] >= 0)
        index_tab[i][j] = (LONG *)m_alloc((n_part - np_tab[i] + 1)
                  * sizeof(LONG), IMSTR);
      else
        index_tab[i][j] = 0;
     }
   }

  for(i=0; i<tab_size; i++)
   for(j=0; j<ysize; j++)
    for(lnp=0; lnp<=n_part - np_tab[i]; lnp++)
     {
      np = lnp + np_tab[i];
      cnt = j * xsize;
      index = 0;
      mask = ((LLONG) 1)<<(xsize-1);
      for(x=0; x<xsize; x++)
       {
        if(i & mask)
         {
          index += over(sites-cnt-1, np);
          np--;
          if(np < 0)
           fprintf(stderr, "### Error: Illegal bit pattern %x encountered:\n", i);
         }
        cnt++;
        mask = mask>>1;
       }
      index_tab[i][j][lnp] = index; 
     }
 }

/* Free mapping tables */

static void free_mtables()
 {
  register int i, j;

  for(i=0; i<tab_size; i++)
   {
    for(j=0; j<ysize; j++)
       if(index_tab[i][j])
           free(index_tab[i][j]);
#ifndef xsize
    free(index_tab[i]);
#endif
   }
#ifndef xsize
  free(index_tab);
#endif
  free(np_tab);
 }
#endif

/* Generate states in S^z-symmetry subspace */

static void gen_states()
 {
  nstates = over(sites, n_part);
 }

/* Free states in S^z-symmetry subspace */

static void free_states()
 {
 }

/* Convert index to bit representation of state */
/* needs pointer to space for state to be passed
   (more efficient than allocating memory for it
    each time this function is called) */

static void index_to_state(LONG index, LLONG *state)
 {
  LONG oindex, hh;
  LLONG mask;
  int np;
  int count;
  register int i, j;

  for(j=0; j<ysize; j++)
    state[j] = 0;

  oindex = index;

  count = 0;
  np = n_part;
  for(j=0; j<ysize; j++)
   {
    mask = ((LLONG) 1)<<(xsize-1);
    for(i=0; i<xsize; i++)
     {
      hh = over(sites-count-1, np);
#ifdef DEBUG
      printf("(%d,%d) = %d\n", sites-count-1, np, hh);
#endif
      if(index >= hh)
       {
        index -= hh;
        np--;
        state[j] |= mask;
       }
      count++;
      mask = mask>>1;
     }
   }

  if(index || (np < 0))
   fprintf(stderr, "*** Error: failed to translate index `%ld' to state\n", oindex);

#ifdef DEBUG
  fprintf(stderr, "Result of index_to_state:\n");
  for(j=0; j<ysize; j++)
     fprintf(stderr, "\t%lx\n", state[j]);
#endif
 }

/* Convert bit representation of state to index of matrix */

#ifndef INDEX_INLINE

static void show_state(FILE *out, LLONG *state);

static void state_to_index(LLONG *state, LONG *index)
 {
#ifdef SLOWINDEX
  LLONG mask;
  int count;
#else
  int lnp;
#endif
  int np;
  register int i, j;
/*
  register int k;
*/

#ifdef SLOWINDEX
  count = *index = 0;
  np = n_part;
  for(j=0; j<ysize; j++)
   {
    mask = ((LLONG) 1)<<(xsize-1);
    for(i=0; i<xsize; i++)
     {
      if(state[j] & mask)
       {
        *index += over(sites-count-1, np);
        np--;
        if(np < 0)
         {
          fprintf(stderr, "### Error: Illegal state encountered:\n");
          show_state(stderr, state);
/*
          for(k=0; k<ysize; k++)
             fprintf(stderr, "\t%lx\n", state[k]);
*/
         }
       }
      count++;
      mask = mask>>1;
     }
   }
#else
  *index = 0;
  np = n_part;
  for(j=0; j<ysize; j++)
   {
    lnp = np-np_tab[state[j]];
    if(lnp < 0)
     {
      fprintf(stderr, "### Error: Illegal state encountered:\n");
      show_state(stderr, state);
/*
      for(k=0; k<ysize; k++)
         fprintf(stderr, "\t%lx\n", state[k]);
*/
     }
    else
     *index += index_tab[state[j]][j][lnp];
    np = lnp;
   }
#endif

#ifdef DEBUG
  fprintf(stderr, "State:\n");
  for(k=0; k<ysize; k++)
     fprintf(stderr, "\t%lx\n", state[k]);
  fprintf(stderr, "Index:%ld\n", *index);
#endif
 }

#else

/* Same thing inline - without error checking !!! */

#define state_to_index(__state, __index) {\
  *__index = 0;\
  __np = n_part;\
  for(__j=0; __j<ysize; __j++) \
    *__index += index_tab[__state[__j]][__j][__np -= np_tab[__state[__j]] ]; }
 
#endif

/* Show a state (subroutine) */

static void show_bitrep(FILE *out, LLONG map)
 {
  register LLONG mask;
  register int i;

  mask = (((LLONG) 1)<<(xsize-1));
  for(i=0; i<xsize; i++)
   {
    fputc((map&mask)?'1':'0', out);
    mask = mask>>1;
   }
 }

/* Show a state */

static void show_state(FILE *out, LLONG *state)
 {
  register int i;

  for(i=0; i<ysize; i++)
   {
    if(i)
     fprintf(out, "\n ");
    else
     fprintf(out, "|");
    show_bitrep(out, state[i]);
   }
  fprintf(out, ">\n");
 }

/* Translate a state bitwise left */

static void translate_left(LLONG *state)
 {
  register LLONG mask;
  register int i;

  mask = (((LLONG) 1)<<xsize) - 1;
  if(! mask)
    mask = -1;
  for(i=0; i<ysize; i++)
    state[i] = (((state[i]<<1) | ((state[i] >> (xsize - 1)) & 1)
               ) & mask);
 }

/* Translate a state in y-direction */

/* *********************
   We can do this inline

static void translate_y(LLONG *state)
 {
  register int k;
  LLONG temp;

  temp = state[0];
  for(k=0; k<ysize-1; k++)
    state[k] = state[k+1];
  state[ysize-1] = temp;
 }

*/

#define translate_y(state) {\
  register int __k; LLONG __temp; \
  __temp = state[0]; \
  for(__k=0; __k<ysize-1; __k++) state[__k] = state[__k+1]; \
  state[ysize-1] = __temp; }

/* ********************* */

/* Reflect state along the x-direction */

static void reflect_x(LLONG *state)
 {
  register int j, x;
  LLONG tempa, tempb;
  LLONG mask, rmask;

  for(j=0; j<ysize; j++)
   {
     tempa = state[j];
     tempb = 0;
     rmask = mask = 1;
     for(x = 0; x<xsize; x++)
      {
       if(tempa & mask)
         tempb |= rmask;
       mask = (mask << 1);
       if(rmask == 1)
         rmask = ((LLONG) 1) << (xsize-1);
       else
         rmask = (rmask >> 1);
      }
     state[j] = tempb;
   }
 }

/* Reflect state along the y-direction */

static void reflect_y(LLONG *state)
 {
  register int j;
  LLONG temp;

  for(j=1; j<=(ysize-1)/2; j++)
   {
     temp = state[j];
     state[j] = state[ysize-j];
     state[ysize-j] = temp;
   }
 }

/* Reflect state along the diagonal */

static void reflect_d(LLONG *state)
 {
  register int x, y;
  register LLONG xmask, ymask;
#ifndef MAXY
  LLONG *hstate;
#else
  LLONG hstate[MAXY];
#endif

#ifndef MAXY
  hstate = (LLONG *) m_alloc(sizeof(LLONG) * ysize, "reflect_d");
#endif

  for(y=0; y<ysize; y++)
   {
    hstate[y] = state[y];
    state[y] = 0;
   }

  ymask = 1;
  for(y=0; y<ysize; y++)
   {
    xmask = 1;
    for(x=0; x<xsize; x++)
     {
      if(hstate[y] & xmask)
        state[x] |= ymask;
      xmask = xmask << 1;
     }
    ymask = ymask<<1;
   }

#ifndef MAXY
  free(hstate);
#endif
 }

/* Create the momentum basis */

static void create_momenta()
 {
#ifdef INDEX_INLINE
  int __np, __j;
#endif
  register long i;
  register int j, k;
  int flag;
  int tcnt;
#ifndef MAXY
  LLONG *state;
#else
  LLONG state[MAXY];
#endif
  int refl, refld;
  int sign, fac;
  LONG index;
  long ps[26000];
  int pcnt;
  SLONG cmi;

  charge_momentum = (SLONG *)m_alloc(sizeof(SLONG)*nstates, "create_momenta");
  charge_prefac = (signed short *)m_alloc(sizeof(signed short)*nstates, "create_momenta");

#ifndef MAXY
  /* Reserve memory for the state */
  state = (LLONG *) m_alloc(sizeof(LLONG) * ysize, "create_momenta");
#endif

#pragma omp parallel for private(i) if(nstates>5000)
  for(i=0; i<nstates; i++)
   {
    charge_momentum[i] = MAGIC;
    charge_prefac[i] = 0;
   }

  pstates = 0;
  for(i=0; i<nstates; i++)
   if(charge_momentum[i] == MAGIC)
    {
     index_to_state(i, state);
     sign = 1;
     pcnt = 0;

     for(refld=0; refld<(parity_d?2:1); refld++)
      {				/* Orbit over diagonal reflections */
     for(refl=0; refl<4; refl++)
      {				/* Orbit over reflection group */
       for(j=0; j<xsize; j++)
        {			/* Orbit over translational group */
         for(tcnt=0; tcnt<ysize; tcnt++)
          {
           state_to_index(state, &index);
           charge_momentum[index] = pstates;   
           charge_prefac[index] += sign;

           for(k=0; (k<pcnt) && (ps[k] != index); k++)
                    ;		/* Remember all states which appeared */
           if(k==pcnt)
              ps[pcnt++] = index;

           translate_y(state);
           sign *= momentum_y;
          }
         translate_left(state);
         sign *= momentum_x;
        }			/* End of translation group */
       if(refl & 1)
        {
         reflect_y(state);
         sign *= parity_y;
        }
       else
        {
         reflect_x(state);
         sign *= parity_x;
        }
      }				/* End of reflection group */
       if(parity_d)
        {
#ifdef DEBUGREFLD
         printf("Before reflection:\n");
         show_state(state);
#endif
         reflect_d(state);
#ifdef DEBUGREFLD
         printf("After reflection:\n");
         show_state(state);
#endif
         sign *= parity_d;
        }
      }				/* End of diagonal reflections */

     fac = -1;			/* Test prefactors */
     for(j=0; j<pcnt; j++)
        {
         sign = charge_prefac[ps[j]];
         if(sign < 0)
           sign = -sign;
         if(sign)
           charge_prefac[ps[j]] /= sign;
         if(fac < 0)
           fac = sign;
         if(fac != sign)
          {
           fprintf(stderr, "### Error: Not all prefactors have equal magnitude\n");
           exit(-15);
          }
        }

     if(! fac)
      {
       for(j=0; j<pcnt; j++)
           charge_momentum[ps[j]] = MAGIX;
      }
     else
       pstates++;
    }

#ifndef MAXY
  free(state);			/* Now we don't need memory for the state any longer */
#endif

  momentum_charge = (LONG *)m_alloc(sizeof(LONG)*pstates, "create_momenta");
  momentum_prefac = (INT *)m_alloc(sizeof(INT)*pstates, "create_momenta");

#pragma omp parallel for private(i) if(pstates>5000)
  for(i=0; i < pstates; i++)
   {
    momentum_charge[i] = MAGIC;
    momentum_prefac[i] = 0; 
   }

  for(i=0; i<nstates; i++)
   if((cmi = charge_momentum[i]) < MAXSNSTATES)
     {
       if(momentum_charge[cmi] >= MAXSNSTATES)
         momentum_charge[cmi] = i;
       momentum_prefac[cmi]++;
     }
 }

/* Apply Hamilton operator */

void apply_internal_ham(struct vector *src, struct vector *dst, char *name)
 {
#ifdef INDEX_INLINE
  int __np, __j;
#endif
  register long i, j;
  register int x;
  int nj;
  LLONG mask, nmask, lmask;
  LLONG tempa, tempb, tempam;
  LONG index;
  int szsz, szsza;
#ifndef MAXY
  LLONG *state;
#else
  LLONG state[MAXY];
#endif
  REAL prefac, jval2, jval_a2;
  LONG mc;
  SLONG sind;
  signed short cpmc;
  COMPLEX result;
  INT mpf;
  REAL lprefac[LBUFSIZ];
  SLONG lsind[LBUFSIZ];
  long lcnt;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  if(dst->el_cnt != pstates)
   {
    display_error("Dimension of output vector doesn't match - can't do internal matrix application");
    return;
   }

#ifndef MAXY
  /* Reserve memory for the state */
  state = (LLONG *) m_alloc(sizeof(LLONG) * ysize, "apply_internal_ham");
#endif

  jval2 = (jval/((REAL) 2));		/* Precompute J/2 */
  jval_a2 = (jval*alpha/((REAL) 2));	/* Precompute J*alpha/2 */

#ifdef INDEX_INLINE
#pragma omp parallel for private(i,j,nj,szsza,szsz,mc,cpmc,mpf,result,\
         lprefac,lsind,lcnt,mask,nmask,lmask,x,\
         tempa,tempb,tempam,state,index,sind,prefac,\
         __np,__j) if(pstates>1000)
#pragma omp parallel for private(i,j,nj,szsza,szsz,mc,cpmc,mpf,result,\
         lprefac,lsind,lcnt,mask,nmask,lmask,x,\
         tempa,tempb,tempam,state,index,sind,prefac) if(pstates>1000)
#endif
  for(i=0; i<pstates; i++)		/* Loop over all momentum states */
   {
    szsza = szsz = 0;
    mc = momentum_charge[i];
    cpmc = charge_prefac[mc];		/* Fetch corresponding charge_prefac */
    mpf = momentum_prefac[i];		/* Fetch momentum_prefac[i] */
    index_to_state(mc, state);
    lcnt = 0;				/* Nothing buffered yet */
    for(j=0; j<ysize; j++)		/* Loop over all sites */
     {
      tempa = state[j];			/* The state at row j is in tempa */
      nj = (j+1) % ysize;		/* next index in y-direction */
      tempb = state[nj];		/* The state at row nj is in tempb */

      mask = 1;				/* Mask for site x */
      lmask = (((LLONG) 1) << (xsize-1));		/* Mask for site xsize-1 */
      for(x=0; x<xsize; x++)
       {
        if(x == (xsize-1))		/* next mask in x-direction */
          nmask = 1;
        else
          nmask = mask<<1;

        tempam = tempa & mask;		/* Store one masked object */

        if( (tempam && (! (tempa & nmask))) ||
            ((! tempam) && (tempa & nmask)) )	/* x-direction */
         {				/* S^+ S^- - interaction */
           state[j] = tempa ^ (mask | nmask);
           state_to_index(state, &index);
           sind = charge_momentum[index];
           if(sind < MAXSNSTATES)
            {
             if(cpmc != charge_prefac[index])
                prefac = -jval2;
             else
                prefac = jval2;
             if(momentum_prefac[sind] != mpf)
                prefac *= sqrt(((REAL) mpf) / ((REAL) momentum_prefac[sind]));
             lprefac[lcnt] = prefac;		/* Store this for later execution */
             lsind[lcnt++] = sind;
            }
           szsz--;
         }
        else
           szsz++;

        state[j] = tempa ^ mask;	/* In the next 3 cases we change row j of state */

        if(tempam != (tempb & mask))	/* y-direction */
         {				/* S^+ S^- - interaction */
          state[nj] = tempb ^ mask;
          state_to_index(state, &index);
          sind = charge_momentum[index];
          if(sind < MAXSNSTATES)
            {
             if(cpmc != charge_prefac[index])
                prefac = -jval2;
             else
                prefac = jval2;
             if(momentum_prefac[sind] != mpf)
                prefac *= sqrt(((REAL) mpf) / ((REAL) momentum_prefac[sind]));
             lprefac[lcnt] = prefac;	/* Store this for later execution */
             lsind[lcnt++] = sind;
            }
          szsz--;
         }
        else
          szsz++;

        if( (tempam && (! (tempb & nmask))) ||	/* The first diagonal */
             ((! tempam) && (tempb & nmask)) )
         {                              /* S^+ S^- - interaction */
          state[nj] = tempb ^ nmask;
          state_to_index(state, &index);
          sind = charge_momentum[index];
          if(sind < MAXSNSTATES)
           {
             if(cpmc != charge_prefac[index])
                prefac = -jval_a2;
             else
                prefac = jval_a2;
             if(momentum_prefac[sind] != mpf)
                prefac *= sqrt(((REAL) mpf) / ((REAL) momentum_prefac[sind]));
             lprefac[lcnt] = prefac;		/* Store this for later execution */
             lsind[lcnt++] = sind;
           }
          szsza--;
         }
        else
          szsza++;

        if( (tempam && (! (tempb & lmask))) ||	/* The second diagonal */
             ((! tempam) && (tempb & lmask)) )
         {                              /* S^+ S^- - interaction */
          state[nj] = tempb ^ lmask;
          state_to_index(state, &index);
          sind = charge_momentum[index];
          if(sind < MAXSNSTATES)
           {
             if(cpmc != charge_prefac[index])
                prefac = -jval_a2;
             else
                prefac = jval_a2;
             if(momentum_prefac[sind] != mpf)
                prefac *= sqrt(((REAL) mpf) / ((REAL) momentum_prefac[sind]));
             lprefac[lcnt] = prefac;	/* Store this for later execution */
             lsind[lcnt++] = sind;
           }
          szsza--;
         }
        else
          szsza++;

        state[nj] = tempb;		/* Now restore state[nj] */

        lmask = mask;			/* Compute masks for next pass */
        mask = nmask;
       }
      state[j] = tempa;			/* Now restore also state[j] */
     }

    /* Now perform buffered multiplications */
    result.re = result.im = 0;
    for(j=0; j<lcnt; j++)
     {
      result.re += (lprefac[j] * src->elements[lsind[j]].re);
      result.im += (lprefac[j] * src->elements[lsind[j]].im);
     }

    prefac = jval*Delta/4.0L*(((REAL) szsz) + alpha*((REAL) szsza));
						/* The diagonal element */

    /* And store final result */
    dst->elements[i].re = result.re + (prefac * src->elements[i].re);
    dst->elements[i].im = result.im + (prefac * src->elements[i].im);
   }

#ifndef MAXY
  free(state);		/* Now we don't need memory for the state any longer */
#endif
 }

/* Routine to compute correlation functions with respect to the vector in src 
   (put an empty one, if you don't need it) */

extern void compute_correl(struct vector *src, FILE *fpo)
 {
#ifdef INDEX_INLINE
  int __np, __j;
#endif
  COMPLEX ctmpa, ctmpb, ctmpc;
  struct vector dst, dstpm;
  int x, y, x0, y0, nx, ny, direc;
  register long i;
  LLONG mask, nmask;
  LLONG tempa, tempb;
  LONG index;
  int szsz;
#ifndef MAXY
  LLONG *state;
#else
  LLONG state[MAXY];
#endif
  REAL prefac;
  LONG mc;
  SLONG sind;
  signed short cpmc;
  COMPLEX resultpm;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  alloc_int_vector(&dst, pstates);
  alloc_int_vector(&dstpm, pstates);

  sprod(src, src, &ctmpa);

  fprintf(fpo, "\nmatrix Szz(%d,%d);\n\n", xsize, ysize);
  fprintf(fpo, "matrix Spm(%d,%d);\n\n", xsize, ysize);

#ifndef MAXY
  /* Reserve memory for the state */
  state = (LLONG *) m_alloc(sizeof(LLONG) * ysize, "compute_correl");
#endif

  for(x = 0; x<xsize; x++)
   for(y = 0; y<ysize; y++)
    {
#ifdef INDEX_INLINE
#pragma omp parallel for private(i,j,nj,szsza,szsz,mc,cpmc,mpf,result,\
         lprefac,lsind,lcnt,mask,nmask,lmask,x,\
         tempa,tempb,tempam,state,index,sind,prefac,\
         __np,__j) if(pstates>1000)
#else
#pragma omp parallel for private(i,szsz,mc,cpmc,resultpm,\
         mask,nmask,x0,y0,nx,ny,direc,\
         tempa,tempb,state,index,sind,prefac) if(pstates>1000)
#endif
     for(i=0; i<pstates; i++)		/* Loop over all momentum states */
      {
       szsz = 0;
       mc = momentum_charge[i];
       cpmc = charge_prefac[mc];	/* Fetch corresponding charge_prefac */
       index_to_state(mc, state);
       resultpm.re = resultpm.im = 0;
       for(x0 = 0; x0<xsize; x0++)	/* Loop over all sites */
        for(y0 = 0; y0<ysize; y0++)
         {
          mask = ((LLONG) 1)<<x0;
          tempa = state[y0];
          for(direc = 0; direc<(parity_d?8:4); direc++)	/* and directions */
          {
           switch(direc)
            {
             case 0:
              nx = x0+x;
              ny = y0+y;
              break;
             case 1:
              nx = x0-x;
              ny = y0+y;
              break;
             case 2:
              nx = x0+x;
              ny = y0-y;
              break;
             case 3:
              nx = x0-x;
              ny = y0-y;
              break;
             case 4:
              nx = x0+y;
              ny = y0+x;
              break;
             case 5:
              nx = x0-y;
              ny = y0+x;
              break;
             case 6:
              nx = x0+y;
              ny = y0-x;
              break;
             case 7:
              nx = x0-y;
              ny = y0-x;
              break;
            }
           while(nx < 0)
             nx += xsize;
           while(nx >= xsize)
             nx -= xsize; 
           while(ny < 0)
             ny += ysize; 
           while(ny >= ysize)
             ny -= ysize; 
           nmask = ((LLONG) 1)<<nx;
           tempb = state[ny];
           if( ((tempa & mask) && (! (tempb & nmask))) ||
                ((! (tempa & mask)) && (tempb & nmask)) )
            {				/* S^+ S^- - correlator */
             state[y0] ^= mask;
             state[ny] ^= nmask;
             state_to_index(state, &index);
             sind = charge_momentum[index];
             if(sind < MAXSNSTATES)
               {
                prefac = (((REAL) 1)/((REAL) ((parity_d?8:4)*2*sites)));
                if(cpmc != charge_prefac[index])
                   prefac = -prefac;
                if(momentum_prefac[sind] != momentum_prefac[i])
                   prefac *= sqrt(((REAL) momentum_prefac[i]) /
                                 ((REAL) momentum_prefac[sind]));
                resultpm.re += (prefac * src->elements[sind].re);
                resultpm.im += (prefac * src->elements[sind].im);
               }
             state[y0] = tempa;
             state[ny] = tempb;
             szsz--;
            }
           else
            {
             if((nx == x0) && (ny == y0))
              {
                prefac = (((REAL) 1)/((REAL) ((parity_d?8:4)*2*sites)));
                resultpm.re += (prefac * src->elements[i].re);
                resultpm.im += (prefac * src->elements[i].im);
              }
             szsz++;
            }
          }
         }
       dstpm.elements[i].re = resultpm.re;
       dstpm.elements[i].im = resultpm.im;
       prefac = ((REAL) szsz)/((REAL) ((parity_d?8:4)*4*sites));
       dst.elements[i].re = (prefac * src->elements[i].re);
       dst.elements[i].im = (prefac * src->elements[i].im);
      }
     sprod(src, &dst, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "Szz(%d,%d) := ", x+1, y+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");

     sprod(src, &dstpm, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "Spm(%d,%d) := ", x+1, y+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");
    }

#ifndef MAXY
  free(state);		/* Now we don't need memory for the state any longer */
#endif

  free_vector(&dstpm);
  free_vector(&dst);
 }

/* Read parameters */

void read_params()
 {
  FILE *fpi;
  char temp_str[256];
  register int i;
  int line;
#ifdef ysize
  int tmp;
#endif

  fpi  = fopen("internal-params", "r");
  if(! fpi)
   {
    fprintf(stderr, "### Error: Failed to open `internal-params' for reading\n");
    exit(-14);
   }
  line = 0;
  while(fgets(temp_str, 255, fpi))
   {
    line ++;
    for(i=0; temp_str[i] && (temp_str[i] <= ' ') && (i < 255); i++)
          ;
    if(temp_str[i] && (temp_str[i] != '#') && (temp_str[i] != '%'))
     {
      if(temp_str[i+1] != '=')
        {
         fprintf(stderr, "### Error: Not a `=' as second character on line %d in file `internal-params'\n", line);
         exit(-14);
         }
       switch(temp_str[i] | 0x20)
         {
           case 'x':
#ifdef xsize
             sscanf(temp_str+i+2, "%d", &tmp);
             if(tmp != xsize)
              {
               fprintf(stderr, "### Error: This version supports only a fixed x=%d\n", xsize);
               exit(-14);
              }
#else
             sscanf(temp_str+i+2, "%d", &xsize);
#endif
             break;
           case 'y':
#ifdef ysize
             sscanf(temp_str+i+2, "%d", &tmp);
             if(tmp != ysize)
              {
               fprintf(stderr, "### Error: This version supports only a fixed y=%d\n", ysize);
               exit(-14);
              }
#else
             sscanf(temp_str+i+2, "%d", &ysize);
#endif
             break;
           case 'n':
             sscanf(temp_str+i+2, "%d", &n_part);
             break;
           case 'k':
             momentum_x = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'l':
             momentum_y = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'r':
             parity_x = (temp_str[i+2] == '-')?-1:1;
             break;
           case 's':
             parity_y = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'd':
             parity_d = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'j':
             sscanf(temp_str+i+2, "%lg", &jval);
             break;
           case 'a':
             sscanf(temp_str+i+2, "%lg", &alpha);
             break;
           case 'z':
             sscanf(temp_str+i+2, "%lg", &Delta);
             break;
           default:
             temp_str[i+1] = 0;
             fprintf(stderr, "### Error: Unknown `%s' as first character on line %d in file `internal-params'\n", temp_str+i, line);
             exit(-14);
         }
     }
   }
  fclose(fpi);

  if((xsize <= 0) || (xsize > MAXD))
   {
    fprintf(stderr, "### Error: x-size is not between 0 and %d\n", MAXD);
    exit(-14);
   }

  if((ysize <= 0) || (ysize > MAXD))
   {
    fprintf(stderr, "### Error: y-size is not between 0 and %d\n", MAXD);
    exit(-14);
   }

   if(parity_d && (xsize != ysize))
    {
     fprintf(stderr, "### Warning: Use reflection along diagonal only for square systems (ignored)\n");
     parity_d = 0;
    }

  if((n_part < 0) || (n_part > xsize * ysize))
   {
    fprintf(stderr, "### Error: Number of particles is not between 0 and %d\n", xsize * ysize);
    exit(-14);
   }

  fprintf(stderr, "\nx-size: %d\ny-size: %d\nnumber of particles: %d\n", xsize, ysize, n_part);

  fprintf(stderr, "J = %1.12lg\n", jval);
  fprintf(stderr, "alpha = %1.12lg\n", alpha);
  fprintf(stderr, "Delta = %1.12lg\n", Delta);
  fprintf(stderr, "momentum x: %s\n", (momentum_x == -1)?"-":"+");
  fprintf(stderr, "momentum y: %s\n", (momentum_y == -1)?"-":"+");
  fprintf(stderr, "x-parity: %s\n", (parity_x == -1)?"-":"+");
  fprintf(stderr, "y-parity: %s\n", (parity_y == -1)?"-":"+");
  if(parity_d)
    fprintf(stderr, "diagonal parity: %s\n", (parity_d == -1)?"-":"+");
 }

/* Init the internal Hamiltonian application */

long init_interal_h()
 {
  register int i;

  fprintf(stderr, "Internal application of the Hamilton-Matrix\n");
  fprintf(stderr, "for a spin-1/2 Heisenberg model on a square lattice\n");
  fprintf(stderr, "with diagonal bonds\n");
  fprintf(stderr, "written December 1998-January 2005 by Andreas Honecker\n");

  read_params();

  sites = ysize * xsize;	/* Compute the volume of the system */

#ifndef DIRECTBIN
  init_binomials();
#endif
#ifndef SLOWINDEX
  init_mtables();
#endif

  gen_states();

  fprintf(stderr, "\nNumber of states with %d down spins on a %d x %d lattice: %ld\n",
                   n_part, xsize, ysize, nstates); 

  if(nstates >= MAXNSTATES)
   {
    fprintf(stderr, "### Error: This exceeds the storage capacity of the LONG data type\n");
    exit(113);
   }

  create_momenta();

  fprintf(stderr, "\nNumber of associated momentum eigenstates: %ld\n",
                   pstates);

  if(pstates >= MAXSNSTATES)
   {
    fprintf(stderr, "### Error: This exceeds the storage capacity of the SLONG data type\n");
    exit(113);
   }

  return(pstates);
 }

/* Exit the internal Hamiltonian application */

void exit_interal_h()
 {
#ifndef SLOWINDEX
  free_mtables();
#endif
#ifndef DIRECTBIN
  free_binomials();
#endif

  free(momentum_prefac);		/* Now free the tables used */
  free(momentum_charge);
  free(charge_prefac);
  free(charge_momentum);
  
  free_states();
 }
