/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "matrix.h"
#include "vectors.h"
#include "apply_h.h"

/* The parameters */

double tval;
double tpval;
double Uval;

int period;

/* The maximal number of states we can handle */

#define MAXNSTATES ( ((unsigned long) 1) << (sizeof(LONG)*8 - 1) ) + \
     ( ( ((unsigned long) 1) << (sizeof(LONG)*8 - 1) ) - 1000)

/* Two special numbers (should be negative and different, otherwise arbitrary) */

#define MAGIC (MAXNSTATES+123)
#define MAGIX (MAXNSTATES+876)

/* The maximal linear extent */

#define MAXD 32

/* The maximal number of terms in basis state */

#define PSIZE 3000

/* The linear size */

static int size=MAGIC;

/* The number of up and down particles */

static int n_part_u = MAGIC;
static int n_part_d = MAGIC;

/* The various parities */

static int momentum=1;		/* momentum */
#ifndef KNZ
 static int parity=0;		/* reflection */
#endif
static int spin_inversion=1;	/* Spin inversion */

/* Structure of a state */

struct state {long up; long down; };

/* The number of states with given number of up and down electrons */

static long nstates;
static long nstates_u;

/* Dimension of the momentum eigenspace */

static long pstates;

/* Mapping tables */

static LONG *charge_momentum;
	/* Which S^z eigenvector belongs to which momentum eigenvector ? */

#ifndef KNZ
static signed char *charge_prefac;
	/* The (sign) prefactors */

#else
static unsigned char *rou;
	/* Which phase factor (root of unity) ? */

static int nrou;	/* The number of roots of unity */
static COMPLEX *rou_val;
#endif

static LONG *momentum_charge;
	/* Representative S^z eigenvector for the momentum eigenvector */

static INT *momentum_prefac;
	/* The number of states associated to the momentum eigenvector */

/* More tables to convert bit representations to index */

static long *bit_to_index;		/* Bit representation to index */
static long **index_to_bit;		/* Index to bit representation for given number of particles */
static signed char *np_tab;		/* number of particles */

/* ****** Some math routines ****** */

/* Den groessten gemeinsamen Teiler zweier Zahlen feststellen
   ueber Euklidischen Algorithmus
   Zehnerpotenzen schneller als Primfaktorzerlegung */

static long get_gcd(long a, long b)
 {
   register int i;
   long remainder;

   if(a == 0)           /* Sonderbehandlung fuer Nullen */
     return(1);
   if(b == 0)
     return(1);

   if(a < 0)            /* Betrag bilden */
     a=-a;
   if(b < 0)
     b=-b;

   if(a == 1)           /* Sonderbehandlung fuer Einsen */
     return(1);
   if(b == 1)
     return(1);

   remainder = a % b;   /* Das ist der Euklische Algorithmus */
   while(remainder)
    {
      a=b;
      b=remainder;
      remainder = a % b;
    }           /* Nun steht der GGT in b */

   return(b);
 }

/* Der Binomialkoeffizient */

static long over(long n, long m)
 {
   register long i;
   register long gcd;
   register long res1, res2;

   if(m > n-m)                  /* Symmetrie ausnutzen */
     m = n-m;

   if(m < 0)
     return(0);

   res1 = 1;
   res2 = 1;
   for(i=0; i<m; i++)
    {
      res1 *= (n-i);           /* Multiplizieren */
      res2 *= (i+1);
      gcd = get_gcd(res1, res2);      /* Und kuerzen */
      res1 /= gcd;
      res2 /= gcd;
    }
   if(res2 != 1)
     fprintf(stderr, "*** Error: Binomial coefficient (%ld,%ld) not an integer !\n", n, m);
   return(res1);
 }

/* Allocate memory and produce error message if not available */

static void *m_alloc(long size, char *str)
 {
  void *ptr;

  ptr = (void *)malloc(size);

  if(! ptr)
   {
    fprintf(stderr, "### Error: Out of memory in %s()\n", str);
    exit(-13);
   }
  return(ptr);
 }

/* Initialize mapping tables */

static char IMSTR[] = "init_mtables";

static void init_mtables()
 {
  long tsize;
  long index, state;
  register unsigned long mask;
  int count;
  int np;
  register int i, j, k;

  tsize = 1<<size;

  bit_to_index = (long *)m_alloc(tsize * sizeof(long), IMSTR);
  index_to_bit = (long **)m_alloc((size+1) * sizeof(long *), IMSTR);
  for(np=0; np<=size; np++)
    index_to_bit[np] = (long *)m_alloc(over(size, np) * sizeof(long), IMSTR);
  np_tab = (signed char *)m_alloc(tsize * sizeof(signed char), IMSTR);

  for(state=0; state<tsize; state++)
   {
    mask = 1;			/* Count the number of bits */
    np = 0;
    for(i=0; i<size; i++)
     {
      if(state & mask)
        np++;
      mask = mask<<1;
     }
    np_tab[state] = np;

    count = index = 0;		/* Convert bit representation to index */
    mask = 1<<(size-1);
    for(i=0; i<size; i++)
     {
      if(state & mask)
       {
        index += over(size-count-1, np);
        np--;
        if(np < 0)
         {
          fprintf(stderr, "### Error: Illegal state encountered: %lx\n", state);
         }
       }
      count++;
      mask = mask>>1;
     }
    bit_to_index[state] = index;		/* Store both directions */
    index_to_bit[np_tab[state]][index] = state;
   }
 }

/* Free mapping tables */

static void free_mtables()
 {
  register int np;

  free(np_tab);

  for(np=0; np<=size; np++)
    free(index_to_bit[np]);
  free(index_to_bit);
  
  free(bit_to_index);
 }

/* Generate states with given number of up and down spins */

static void gen_states()
 {
  nstates_u = over(size, n_part_u);
  nstates = nstates_u * over(size, n_part_d);
 }

/* Free states with given number of up and down spins */

static void free_states()
 {
 }

/* Convert index to bit representation of state */

/* ******************************
   This is also easy to do inline

static void index_to_state(long index, struct state *state)
 {
  long uindex, dindex;

  uindex = index % nstates_u;
  dindex = index / nstates_u;

  state->up = index_to_bit[n_part_u][uindex];
  state->down = index_to_bit[n_part_d][dindex];
 }

*/

#define index_to_state(index, state) {\
    (state)->up = index_to_bit[n_part_u][index % nstates_u];\
    (state)->down = index_to_bit[n_part_d][index / nstates_u]; }

/* ****************************** */

/* Convert bit representation of state to index of matrix */

#define state_to_index(state) (bit_to_index[(state)->down] * nstates_u + bit_to_index[(state)->up])

/* Show a state (subroutine) */

static show_bitrep(long map)
 {
  register unsigned long mask;
  register int i;

  mask = (1<<(size-1));
  for(i=0; i<size; i++)
   {
    putchar((map&mask)?'1':'0');
    mask = mask>>1;
   }
 }

/* Show a state */

static show_state(struct state *state)
 {
  putchar('|');
  show_bitrep(state->up);
  putchar(',');
  show_bitrep(state->down);
  putchar('>');
 }

/* Translate a state bitwise left */

static int translate_left(struct state *state)
 {
  register unsigned long mask;
  register int sign;

  mask = ((unsigned long) (1<<size)) - 1;
  sign = 1;
  if(! mask)
    mask = -1;
  if((state->up & 1) && (! (n_part_u & 1)))	/* Fermionic sign */
    sign = -sign;
  state->up = (((state->up<<1) | ((state->up >> (size - 1)) & 1)) & mask);
  if((state->down & 1) && (! (n_part_d & 1)))	/* Fermionic sign */
    sign = -sign;
  state->down = (((state->down<<1) | ((state->down >> (size - 1)) & 1)) & mask);
  return(sign);
 }

/* Reflect state */

#ifndef KNZ
static void reflect(struct state *state)
 {
  register int x;
  unsigned long tempu, tempd;
  unsigned long mask, rmask;

  tempd = tempu = 0;
  rmask = 1;
  mask = 2;
  for(x = 0; x<size; x++)
   {
    if(state->up & mask)
      tempu |= rmask;
    if(state->down & mask)
      tempd |= rmask;
    mask = (mask << 1);
    if(mask == (1<<size))
      mask = 1;
    if(rmask == 1)
      rmask = 1 << (size-1);
    else
      rmask = (rmask >> 1);
   }
  state->up = tempu;
  state->down = tempd;
 }
#endif

/* Spin inversion */

static void s_invert(struct state *state)
 {
  register unsigned long temp;

  temp = state->up;
  state->up = state->down;
  state->down = temp;
 }

/* Simplify roots */

#ifdef KNZ
void simplify_roots(int *rou_list)
 {
  register int i, j;
  int cnt;
  int val, flag;

  if(! (nrou % 5))	/* 5th roots */
   {
    cnt = nrou/5;
    for(i=0; i<nrou-cnt; i++)
     if(rou_list[i])
      {
       flag = TRUE;
       val = rou_list[i+cnt];
       for(j=2; j<5; j++)
         if(rou_list[(i+j*cnt) % nrou] != val)
              flag = FALSE;
       if(flag)
        {
          rou_list[i] -= val;
          for(j=1; j<5; j++)
            rou_list[(i+j*cnt) % nrou] = 0;
        }
      }
   }

  if(! (nrou % 3))	/* 3rd roots */
   {
    cnt = nrou/3;
    for(i=0; i<cnt; i++)
     {
      if(rou_list[i] && (rou_list[i] == rou_list[i+cnt]))
       {
         rou_list[i+cnt+cnt] -= rou_list[i];
         rou_list[i] = rou_list[i+cnt] = 0;
       } 
      if(rou_list[i] && (rou_list[i] == rou_list[i+cnt+cnt]))
       {
         rou_list[i+cnt] -= rou_list[i];
         rou_list[i] = rou_list[i+cnt+cnt] = 0;
       }
      if(rou_list[i+cnt] && (rou_list[i+cnt] == rou_list[i+cnt+cnt]))
       {
         rou_list[i] -= rou_list[i+cnt];
         rou_list[i+cnt] = rou_list[i+cnt+cnt] = 0;
       }
     }
   }

  if(! (nrou & 1))	/* 2nd roots */
   {
    cnt = nrou/2;
    for(i=0; i<cnt; i++)
     {
      rou_list[i] -= rou_list[i+cnt];
      rou_list[i+cnt] = 0;
     }
   }
 }
#endif

/* Create the momentum basis */

static void create_momenta()
 {
  register long i;
  register int j, k;
#ifdef KNZ
  register int l;
#endif
  int flag;
  int tcnt;
  struct state state;
#ifndef KNZ
  int refl, sign;
#else
  signed int root;
#endif
  int sinv, fac, val;
  long index;
  long ps[PSIZE];
  int pcnt;
#ifdef KNZ
  int **rou_list;
#endif

#ifdef KNZ
  nrou = size/period;
  if(nrou & 1)
    nrou *= 2;
  rou_val = (COMPLEX *)m_alloc(sizeof(COMPLEX)*nrou, "create_momenta");
  for(i=0; i<nrou; i++)
   {
    rou_val[i].re = cos(2.0*PI*i/((double) nrou));
    rou_val[i].im = sin(2.0*PI*i/((double) nrou));
   }
 #ifdef DEBUGKNZ
  for(i=0; i<nrou; i++)
   {
    printf("rou_val[%d] = ", i);
    show_complex(rou_val+i);
   }
 #endif
#endif

  charge_momentum = (LONG *)m_alloc(sizeof(LONG)*nstates, "create_momenta");
#ifndef KNZ
  charge_prefac = (signed char *)m_alloc(sizeof(signed char)*nstates, "create_momenta");
#else
  rou = (unsigned char *)m_alloc(sizeof(unsigned char)*nstates, "create_momenta");
#endif

  for(i=0; i<nstates; i++)
   {
    charge_momentum[i] = MAGIC;
#ifndef KNZ
    charge_prefac[i] = 0;
#else
    rou[i] = 0;
#endif
   }

#ifdef KNZ
  rou_list = (int **)m_alloc(sizeof(int *)*PSIZE, "create_momenta");
  for(i=0; i<PSIZE; i++)
    rou_list[i] = 0;
#endif

  pstates = 0;
  for(i=0; i<nstates; i++)
   if(charge_momentum[i] == MAGIC)
    {
     index_to_state(i, &state);
#ifndef KNZ
     sign = 1;
#else
     root = 0;
#endif
     pcnt = 0;

     for(sinv=0; sinv<=((n_part_u == n_part_d)?1:0); sinv++)
      {				/* Orbit over spin inversions */
#ifndef KNZ
     for(refl=0; refl<=((parity)?1:0); refl++)
      {				/* Orbit over reflection group */
#endif
#ifdef DEBUGTRANS
 #ifndef KNZ
       printf("%d*", sign);
 #else
       printf("[%d]*", root);
 #endif
       show_state(&state);
#endif
       for(j=0; j<size/period; j++)
        {			/* Orbit over translational group */
#ifdef DEBUGTRANS
 #ifndef KNZ
         printf(" -> %d*", sign);
 #else
         printf(" -> [%d]*", root);
 #endif
         show_state(&state);
         printf("\n");
#endif
         index = state_to_index(&state);
         charge_momentum[index] = pstates;   

         for(k=0; (k<pcnt) && (ps[k] != index); k++)		/* Remember all states which appeared */
                  ;		/* ... first find them */

#ifndef KNZ
         charge_prefac[index] += sign;
#else
         if(! rou_list[k])
          {
           rou_list[k] = (int *)m_alloc(sizeof(int)*nrou, "create_momenta");
           for(l=0; l<nrou; l++)
             rou_list[k][l] = 0;
          }
         rou_list[k][root]++;
#endif

         if(k==pcnt)		/* ... then store them */
            ps[pcnt++] = index;

#ifndef KNZ
         for(k=0; k<period; k++)
           sign *= translate_left(&state);
         sign *= momentum;
#else
         for(k=0; k<period; k++)
          {			/* Translate left and add fermionic sign factor */
           if(translate_left(&state) == -1)
             root = ((root + nrou/2) % nrou);
          }
         if(nrou*period != size)	/* Get the correct next root */
           root = ((root + 2*momentum) % nrou);
         else
           root = ((root + momentum) % nrou);
#endif
        }			/* End of translation group */
#ifdef DEBUGTRANS
 #ifndef KNZ
       printf("  -->  %d*", sign);
 #else
       printf("  -->  [%d]*", root);
 #endif
       show_state(&state);
       printf("\n");
#endif
#ifndef KNZ
       if(parity)
        {
         reflect(&state);
         sign *= parity;
        }
      }				/* End of reflection group */
#endif
       s_invert(&state);
#ifndef KNZ
       sign *= spin_inversion;
#else
       if(spin_inversion == -1)
         root = ((root + nrou/2) % nrou);
#endif
      }				/* End of spin inversion */

     fac = -1;			/* Test prefactors */
     for(j=0; j<pcnt; j++)
        {
#ifdef KNZ
 #ifdef DEBUGKNZ
         printf("Roots before simplify:");
         for(k=0; k<nrou; k++)
           if(rou_list[j][k])
             printf(" %d*[%d]", rou_list[j][k], k);
         printf("\n");
 #endif
         simplify_roots(rou_list[j]);
 #ifdef DEBUGKNZ
         printf("Roots after simplify:");
         for(k=0; k<nrou; k++)
           if(rou_list[j][k])
             printf(" %d*[%d]", rou_list[j][k], k);
         printf("\n");
 #endif
         val = 0;
         for(k=0; k<nrou; k++)
          {
           if(rou_list[j][k])
            {
             if(val)
              {
               fprintf(stderr, "### Error: Simplification to single root of unity failed\n");
               for(k=0; k<nrou; k++)
                 if(rou_list[j][k])
                   fprintf(stderr, " %d*[%d]", rou_list[j][k], k);
               fprintf(stderr, "\n");
               exit(-27);
              }
             val = rou_list[j][k];
             rou[ps[j]] = k;
            }
          }
#else
         val = charge_prefac[ps[j]];
#endif
         if(val < 0)
          {
           val = -val;
#ifdef KNZ
           rou[ps[j]] = ((rou[ps[j]] + nrou/2) % nrou);
#endif
          }
#ifndef KNZ
         if(val)
           charge_prefac[ps[j]] /= val;
#endif
         if(fac < 0)
           fac = val;
         if(fac != val)
          {
           fprintf(stderr, "### Error: Not all prefactors have equal magnitude\n");
           exit(-15);
          }
        }

     if(! fac)
      {
       for(j=0; j<pcnt; j++)
           charge_momentum[ps[j]] = MAGIX;
      }
     else
       pstates++;
#ifdef KNZ
     for(k=0; k<PSIZE; k++)
     if(rou_list[k])
      {
       free(rou_list[k]);
       rou_list[k] = 0;
      }
#endif
    }

#ifdef KNZ
  free(rou_list);
#endif

  momentum_charge = (LONG *)m_alloc(sizeof(LONG)*pstates, "create_momenta");
  momentum_prefac = (INT *)m_alloc(sizeof(INT)*pstates, "create_momenta");

  for(i=0; i < pstates; i++)
   {
    momentum_charge[i] = MAGIC;
    momentum_prefac[i] = 0; 
   }

  for(i=0; i<nstates; i++)
   if(charge_momentum[i] < MAXNSTATES)
     {
       if(momentum_charge[charge_momentum[i]] >= MAXNSTATES)
         momentum_charge[charge_momentum[i]] = i;
       momentum_prefac[charge_momentum[i]]++;
     }
 }

/* Apply Current operator */

static void apply_current_operator(struct vector *src, struct vector *dst)
 {
  register long i, j;
  register int x;
  int nj;
  unsigned long mask, nmask;
  unsigned long temp;
  long index;
  struct state state;
  REAL prefac;
#ifdef KNZ
  signed int root;
  COMPLEX tempc;
#endif
  LONG mc, dind;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  if(dst->el_cnt != pstates)
   {
    display_error("Dimension of output vector doesn't match - can't do internal matrix application");
    return;
   }

  for(i=0; i<dst->el_cnt; i++)		/* First set destination to zero */
   {
     dst->elements[i].im = dst->elements[i].re = 0.0;
   }

  for(i=0; i<pstates; i++)		/* Loop over all momentum states */
   {
    mc = momentum_charge[i];
    index_to_state(mc, &state);
    mask = 1;
    for(x=0; x<size; x++)		/* Loop over all sites */
     {
      nmask = mask<<1;			/* Nearest neighbour interaction */
      if(x == (size-1))
        nmask = 1;
      temp = state.up;
      if( ((temp & mask) && (! (temp & nmask))) ||
          ((! (temp & mask)) && (temp & nmask)) )
       {				/* hopping of up electrons */
         state.up ^= (mask | nmask);
         dind = charge_momentum[index = state_to_index(&state)];
         if(dind < MAXNSTATES)
          {
           prefac = -((x % period)?tval:tpval);
           if((x == 0) && (! (n_part_u & 1)))	/* Fermionic sign */
               prefac = -prefac;
           if(temp & mask)			/* Hop right has the other sign in current operator */
               prefac = -prefac;
#ifndef KNZ
           if(charge_prefac[mc] != charge_prefac[index])
              prefac = -prefac;
#else
           root = ((signed int) rou[mc]) - ((signed int) rou[index]);
 #ifdef SHOW_MEL
           printf("root = %d\n", root);
 #endif
           while(root < 0)
             root += nrou;
 #ifdef SHOW_MEL
           printf("root = %d\n", root);
 #endif
#endif
           if(momentum_prefac[dind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[dind]));
#ifdef SHOW_MEL
 #ifndef KNZ
           printf("%d -> %d: %1.12lg\n", i, dind, prefac);
 #else
           printf("%d -> %d: %1.12lg * ", i, dind, prefac);
           show_complex(rou_val+root);
 #endif
#endif
           /* Multiply with i*prefac */
#ifndef KNZ
           dst->elements[dind].re -= (prefac * src->elements[i].im);
           dst->elements[dind].im += (prefac * src->elements[i].re);
#else
           multiply(src->elements+i, rou_val+root, &tempc);
           dst->elements[dind].re -= (prefac * tempc.im);
           dst->elements[dind].im += (prefac * tempc.re);
#endif
          }
         state.up = temp;
       }
      temp = state.down;
      if( ((temp & mask) && (! (temp & nmask))) ||
          ((! (temp & mask)) && (temp & nmask)) )
       {                                /* hopping of down electrons */
         state.down ^= (mask | nmask);
         dind = charge_momentum[index = state_to_index(&state)];
         if(dind < MAXNSTATES)
          {
           prefac = -((x % period)?tval:tpval);
           if((x == 0) && (! (n_part_d & 1)))	/* Fermionic sign */
               prefac = -prefac;
           if(temp & mask)			/* Hop right has the other sign in current operator */
               prefac = -prefac;
#ifndef KNZ
           if(charge_prefac[mc] != charge_prefac[index])
              prefac = -prefac;
#else
           root = ((signed int) rou[mc]) - ((signed int) rou[index]);
 #ifdef SHOW_MEL
           printf("root = %d\n", root);
 #endif
           while(root < 0)
             root += nrou;
 #ifdef SHOW_MEL
           printf("root = %d\n", root);
 #endif
#endif
           if(momentum_prefac[dind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[dind]));
#ifdef SHOW_MEL
 #ifndef KNZ
           printf("%d -> %d: %1.12lg\n", i, dind, prefac);
 #else
           printf("%d -> %d: %1.12lg * ", i, dind, prefac);
           show_complex(rou_val+root);
 #endif
#endif
           /* Multiply with i*prefac */
#ifndef KNZ
           dst->elements[dind].re -= (prefac * src->elements[i].im);
           dst->elements[dind].im += (prefac * src->elements[i].re);
#else
           multiply(src->elements+i, rou_val+root, &tempc);
           dst->elements[dind].re -= (prefac * tempc.im);
           dst->elements[dind].im += (prefac * tempc.re);
#endif
          }
         state.down = temp;
       }
      mask = nmask;
     }
   }
 }

/* Apply Hamilton or current operator */
/* (if name starts with "J" or "j", electric current is applied,
    otherwise the Hamiltonian) */

void apply_internal_ham(struct vector *src, struct vector *dst, char *name)
 {
  register long i, j;
  register int x;
  int nj;
  unsigned long mask, nmask;
  unsigned long temp;
  long index;
  int ucnt;
  struct state state;
  REAL prefac;
#ifdef KNZ
  signed int root;
  COMPLEX tempc;
#endif
  LONG mc, dind;

  /* Is it current operator ? */
  if((*name | 0x20) == 'j')
   {
    apply_current_operator(src, dst);
    return;
   }

  /* Otherwise apply Hamiltonian */
  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  if(dst->el_cnt != pstates)
   {
    display_error("Dimension of output vector doesn't match - can't do internal matrix application");
    return;
   }

  for(i=0; i<dst->el_cnt; i++)		/* First set destination to zero */
   {
     dst->elements[i].im = dst->elements[i].re = 0.0;
   }

  for(i=0; i<pstates; i++)		/* Loop over all momentum states */
   {
    ucnt = 0;
    mc = momentum_charge[i];
    index_to_state(mc, &state);

    mask = 1;
    for(x=0; x<size; x++)		/* Loop over all sites */
     {
      nmask = mask<<1;			/* Nearest neighbour interaction */
      if(x == (size-1))
        nmask = 1;
      if((state.up & mask) && (state.down & mask))
         ucnt++;
      temp = state.up;
      if( ((temp & mask) && (! (temp & nmask))) ||
          ((! (temp & mask)) && (temp & nmask)) )
       {				/* hopping of up electrons */
         state.up ^= (mask | nmask);
         dind = charge_momentum[index = state_to_index(&state)];
         if(dind < MAXNSTATES)
          {
           prefac = -((x % period)?tval:tpval);
           if((x == 0) && (! (n_part_u & 1)))	/* Fermionic sign */
               prefac = -prefac;
#ifndef KNZ
           if(charge_prefac[mc] != charge_prefac[index])
              prefac = -prefac;
#else
           root = ((signed int) rou[mc]) - ((signed int) rou[index]);
 #ifdef SHOW_MEL
           printf("root = %d\n", root);
 #endif
           while(root < 0)
             root += nrou;
 #ifdef SHOW_MEL
           printf("root = %d\n", root);
 #endif
#endif
           if(momentum_prefac[dind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[dind]));
#ifdef SHOW_MEL
 #ifndef KNZ
           printf("%d -> %d: %1.12lg\n", i, dind, prefac);
 #else
           printf("%d -> %d: %1.12lg * ", i, dind, prefac);
           show_complex(rou_val+root);
 #endif
#endif
#ifndef KNZ
           dst->elements[dind].re += (prefac * src->elements[i].re);
           dst->elements[dind].im += (prefac * src->elements[i].im);
#else
           multiply(src->elements+i, rou_val+root, &tempc);
           dst->elements[dind].re += (prefac * tempc.re);
           dst->elements[dind].im += (prefac * tempc.im);
#endif
          }
         state.up = temp;
       }
      temp = state.down;
      if( ((temp & mask) && (! (temp & nmask))) ||
          ((! (temp & mask)) && (temp & nmask)) )
       {                                /* hopping of down electrons */
         state.down ^= (mask | nmask);
         dind = charge_momentum[index = state_to_index(&state)];
         if(dind < MAXNSTATES)
          {
           prefac = -((x % period)?tval:tpval);
           if((x == 0) && (! (n_part_d & 1)))	/* Fermionic sign */
               prefac = -prefac;
#ifndef KNZ
           if(charge_prefac[mc] != charge_prefac[index])
              prefac = -prefac;
#else
           root = ((signed int) rou[mc]) - ((signed int) rou[index]);
 #ifdef SHOW_MEL
           printf("root = %d\n", root);
 #endif
           while(root < 0)
             root += nrou;
 #ifdef SHOW_MEL
           printf("root = %d\n", root);
 #endif
#endif
           if(momentum_prefac[dind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[dind]));
#ifdef SHOW_MEL
 #ifndef KNZ
           printf("%d -> %d: %1.12lg\n", i, dind, prefac);
 #else
           printf("%d -> %d: %1.12lg * ", i, dind, prefac);
           show_complex(rou_val+root);
 #endif
#endif
#ifndef KNZ
           dst->elements[dind].re += (prefac * src->elements[i].re);
           dst->elements[dind].im += (prefac * src->elements[i].im);
#else
           multiply(src->elements+i, rou_val+root, &tempc);
           dst->elements[dind].re += (prefac * tempc.re);
           dst->elements[dind].im += (prefac * tempc.im);
#endif
          }
         state.down = temp;
       }
      mask = nmask;
     }

    prefac = Uval * ((REAL) ucnt);		/* The diagonal element */
#ifdef SHOW_MEL
    printf("%d -> %d: %1.12lg\n", i, i, prefac);
#endif
    dst->elements[i].re += (prefac * src->elements[i].re);
    dst->elements[i].im += (prefac * src->elements[i].im);
   }
 }

/* Routine to compute correlation functions with respect to the vector in src
   (put an empty one, if you don't need it) */

void compute_correl(struct vector *src, FILE *fpo)
 {
  COMPLEX ctmpa, ctmpb, ctmpc;
  struct vector dstuu, dstud, dstdd;
  int x, x0, nx, x1, nx1, xz, tmp;
  int sign;
#ifndef KNZ
  int direc;
#endif
  register long i;
  unsigned long mask, nmask, mask1, nmask1, tmask;
  long tempu, tempd;
  long index;
  int nuu, nud, ndd;
  struct state state;
  REAL prefac;
#ifdef KNZ
  signed int root;
  COMPLEX tempc;
#endif
  LONG mc, dind;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  alloc_int_vector(&dstuu, pstates);
  alloc_int_vector(&dstud, pstates);
  alloc_int_vector(&dstdd, pstates);

  sprod(src, src, &ctmpa);		/* Store normalization of source state */

  /* Density-density correlations (these are diagonal bosonic operators) */

  fprintf(fpo, "\nvector Nuu(%d);\n", size);
  fprintf(fpo, "vector Nud(%d);\n", size);
  fprintf(fpo, "vector Ndd(%d);\n\n", size);

  for(x=0; x<size; x++)
   {
     for(i=0; i<pstates; i++)		/* Loop over all momentum states */
      {
       nuu = nud = ndd = 0;
       mc = momentum_charge[i];
       index_to_state(mc, &state);
       for(x0 = 0; x0<size; x0++)	/* Loop over all sites */
        {
          mask = 1<<x0;
#ifndef KNZ
          for(direc = 0; direc<2; direc++)	/* and directions */
           {
           switch(direc)
            {
             case 0:
#endif
              nx = x0+x;
#ifndef KNZ
              break;
             case 1:
              nx = x0-x;
              break;
            }
#endif
          while(nx < 0)
             nx += size;
          while(nx >= size)
             nx -= size;
          nmask = 1<<nx;
          if((state.up & mask) && (state.up & nmask))
            nuu++;
          if((state.up & mask) && (state.down & nmask))
            nud++;
          if((state.down & mask) && (state.down & nmask))
            ndd++;
#ifndef KNZ
           }
#endif
        }
#ifdef KNZ
       prefac = ((REAL) 1)/((REAL) size);
#else
       prefac = ((REAL) 1)/((REAL) 2*size);
#endif
       dstuu.elements[i].re = ((REAL) nuu)*(prefac * src->elements[i].re);
       dstuu.elements[i].im = ((REAL) nuu)*(prefac * src->elements[i].im);
       dstud.elements[i].re = ((REAL) nud)*(prefac * src->elements[i].re);
       dstud.elements[i].im = ((REAL) nud)*(prefac * src->elements[i].im);
       dstdd.elements[i].re = ((REAL) ndd)*(prefac * src->elements[i].re);
       dstdd.elements[i].im = ((REAL) ndd)*(prefac * src->elements[i].im);
      }
     sprod(src, &dstuu, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "Nuu(%d) := ", x+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");

     sprod(src, &dstud, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "Nud(%d) := ", x+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");

     sprod(src, &dstdd, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "Ndd(%d) := ", x+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");
   }

  /* Electron-electron correlations (these are now off-diagonal) */

  fprintf(fpo, "\nvector Cuu(%d);\n", size);
  fprintf(fpo, "vector Cdd(%d);\n\n", size);

  for(x=0; x<size; x++)
   {
     for(i=0; i<dstuu.el_cnt; i++)	/* Vectors must be initialized to zero */
         dstuu.elements[i].im = dstuu.elements[i].re = 0.0;
     for(i=0; i<dstdd.el_cnt; i++)
         dstdd.elements[i].im = dstdd.elements[i].re = 0.0;
     for(i=0; i<pstates; i++)		/* Loop over all momentum states */
      {
       mc = momentum_charge[i];
       index_to_state(mc, &state);
       for(x0 = 0; x0<size; x0++)       /* Loop over all sites */
        {
          mask = 1<<x0;
#ifndef KNZ
          for(direc = 0; direc<2; direc++)      /* and directions */
           {
           switch(direc)
            {
             case 0:
#endif
              nx = x0+x;
#ifndef KNZ
              break;
             case 1:
              nx = x0-x;
              break;
            }
#endif
          while(nx < 0)
             nx += size;
          while(nx >= size)
             nx -= size;
          nmask = 1<<nx;
          tempu = state.up;
          if(state.up & mask)
           {
#ifdef KNZ
             prefac = ((REAL) 1)/((REAL) size);
#else
             prefac = ((REAL) 1)/((REAL) 2*size);
#endif
             for(tmask = (1<<(size-1)); tmask > mask; tmask = (tmask>>1))
               if(state.up & tmask)	/* Accumulate fermionic signs */
                 prefac = -prefac;
             state.up ^= mask;
             if(! (state.up & nmask))
              {
               for(tmask = (1<<(size-1)); tmask > nmask; tmask = (tmask>>1))
                 if(state.up & tmask)	/* Accumulate fermionic signs */
                   prefac = -prefac;
               state.up ^= nmask;
               dind = charge_momentum[index = state_to_index(&state)];
               if(dind < MAXNSTATES)
                {
#ifndef KNZ
                 if(charge_prefac[mc] != charge_prefac[index])
                   prefac = -prefac;
#else
                 root = ((signed int) rou[mc]) - ((signed int) rou[index]);
                 while(root < 0)
                   root += nrou;
#endif
                 if(momentum_prefac[dind] != momentum_prefac[i])
                    prefac *= sqrt(((REAL) momentum_prefac[i]) /
                                   ((REAL) momentum_prefac[dind]));
 
#ifndef KNZ
                 dstuu.elements[dind].re += (prefac * src->elements[i].re);
                 dstuu.elements[dind].im += (prefac * src->elements[i].im);
#else
                 multiply(src->elements+i, rou_val+root, &tempc);
                 dstuu.elements[dind].re += (prefac * tempc.re);
                 dstuu.elements[dind].im += (prefac * tempc.im);
#endif
                }
              }
             state.up = tempu;
           }
          tempd = state.down;
          if(state.down & mask)
           {
#ifdef KNZ
             prefac = ((REAL) 1)/((REAL) size);
#else
             prefac = ((REAL) 1)/((REAL) 2*size);
#endif
             for(tmask = (1<<(size-1)); tmask > mask; tmask = (tmask>>1))
               if(state.down & tmask)	/* Accumulate fermionic signs */
                 prefac = -prefac;
             state.down ^= mask;
             if(! (state.down & nmask))
              {
               for(tmask = (1<<(size-1)); tmask > nmask; tmask = (tmask>>1))
                 if(state.down & tmask)	/* Accumulate fermionic signs */
                   prefac = -prefac;
               state.down ^= nmask;
               dind = charge_momentum[index = state_to_index(&state)];
               if(dind < MAXNSTATES)
                {
#ifndef KNZ
                 if(charge_prefac[mc] != charge_prefac[index])
                   prefac = -prefac;
#else
                 root = ((signed int) rou[mc]) - ((signed int) rou[index]);
                 while(root < 0)
                   root += nrou;
#endif
                 if(momentum_prefac[dind] != momentum_prefac[i])
                    prefac *= sqrt(((REAL) momentum_prefac[i]) /
                                   ((REAL) momentum_prefac[dind]));

#ifndef KNZ
                 dstdd.elements[dind].re += (prefac * src->elements[i].re);
                 dstdd.elements[dind].im += (prefac * src->elements[i].im);
#else
                 multiply(src->elements+i, rou_val+root, &tempc);
                 dstdd.elements[dind].re += (prefac * tempc.re);
                 dstdd.elements[dind].im += (prefac * tempc.im);
#endif
                }
              }
             state.down = tempd;
           }
#ifndef KNZ
           }
#endif
        }
      }
     sprod(src, &dstuu, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "Cuu(%d) := ", x+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");

     sprod(src, &dstdd, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "Cdd(%d) := ", x+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");
   }

  /* Superconducting correlations */

  fprintf(fpo, "\nvector SCuu(%d);\n", size);
  fprintf(fpo, "vector SCdd(%d);\n\n", size);

  for(x=0; x<size; x++)
   {
     for(i=0; i<dstuu.el_cnt; i++)	/* Vectors must be initialized to zero */
         dstuu.elements[i].im = dstuu.elements[i].re = 0.0;
     for(i=0; i<dstdd.el_cnt; i++)
         dstdd.elements[i].im = dstdd.elements[i].re = 0.0;
     for(i=0; i<pstates; i++)		/* Loop over all momentum states */
      {
       mc = momentum_charge[i];
       index_to_state(mc, &state);
       for(xz = 0; xz<size; xz++)       /* Loop over all sites */
        {
#ifndef KNZ
          for(direc = 0; direc<2; direc++)      /* and directions */
           {
           switch(direc)
            {
             case 0:
#endif
              x0 = xz;
              x1 = x0+1;
              nx = x0+x;
              nx1 = x0+x+1;
#ifndef KNZ
              break;
             case 1:
              x0 = xz;
              x1 = x0-1;
              nx = x0-x;
              nx1 = x0-x-1;
              break;
            }
#endif
          while(x1 < 0)
             x1 += size;
          while(x1 >= size)
             x1 -= size;
          while(nx < 0)
             nx += size;
          while(nx >= size)
             nx -= size;
          while(nx1 < 0)
             nx1 += size;
          while(nx1 >= size)
             nx1 -= size;

          sign = -1;
          if(x1 < x0)			/* Order: x1 > x0 */
           {
            tmp = x1;
            x1 = x0;
            x0 = tmp;
            sign = -sign;
           }
          if(nx1 < nx)			/* Order: nx1 > nx */
           {
            tmp = nx1;
            nx1 = nx;
            nx = tmp;
            sign = -sign;
           }

          mask = 1<<x0;			/* Now set up masks */
          mask1 = 1<<x1;
          nmask = 1<<nx;
          nmask1 = 1<<nx1;

          tempu = state.up;
          if((state.up & mask) && (state.up & mask1))
           {
#ifdef KNZ
             prefac = ((REAL) sign)/((REAL) size);
#else
             prefac = ((REAL) sign)/((REAL) 2*size);
#endif
             for(tmask = (mask1>>1); tmask > mask; tmask = (tmask>>1))
               if(state.up & tmask)	/* Accumulate fermionic signs */
                 prefac = -prefac;
             state.up ^= (mask | mask1);
             if((! (state.up & nmask)) && (! (state.up & nmask1)))
              {
               for(tmask = (nmask1>>1); tmask > nmask; tmask = (tmask>>1))
                 if(state.up & tmask)	/* Accumulate fermionic signs */
                   prefac = -prefac;
               state.up ^= (nmask | nmask1);
               dind = charge_momentum[index = state_to_index(&state)];
               if(dind < MAXNSTATES)
                {
#ifndef KNZ
                 if(charge_prefac[mc] != charge_prefac[index])
                   prefac = -prefac;
#else
                 root = ((signed int) rou[mc]) - ((signed int) rou[index]);
                 while(root < 0)
                   root += nrou;
#endif
                 if(momentum_prefac[dind] != momentum_prefac[i])
                    prefac *= sqrt(((REAL) momentum_prefac[i]) /
                                   ((REAL) momentum_prefac[dind]));
 
#ifndef KNZ
                 dstuu.elements[dind].re += (prefac * src->elements[i].re);
                 dstuu.elements[dind].im += (prefac * src->elements[i].im);
#else
                 multiply(src->elements+i, rou_val+root, &tempc);
                 dstuu.elements[dind].re += (prefac * tempc.re);
                 dstuu.elements[dind].im += (prefac * tempc.im);
#endif
                }
              }
             state.up = tempu;
           }

          tempd = state.down;
          if((state.down & mask) && (state.down & mask1))
           {
#ifdef KNZ
             prefac = ((REAL) sign)/((REAL) size);
#else
             prefac = ((REAL) sign)/((REAL) 2*size);
#endif
             for(tmask = (mask1>>1); tmask > mask; tmask = (tmask>>1))
               if(state.down & tmask)	/* Accumulate fermionic signs */
                 prefac = -prefac;
             state.down ^= (mask | mask1);
             if((! (state.down & nmask)) && (! (state.down & nmask1)))
              {
               for(tmask = (nmask1>>1); tmask > nmask; tmask = (tmask>>1))
                 if(state.down & tmask)	/* Accumulate fermionic signs */
                   prefac = -prefac;
               state.down ^= (nmask | nmask1);
               dind = charge_momentum[index = state_to_index(&state)];
               if(dind < MAXNSTATES)
                {
#ifndef KNZ
                 if(charge_prefac[mc] != charge_prefac[index])
                   prefac = -prefac;
#else
                 root = ((signed int) rou[mc]) - ((signed int) rou[index]);
                 while(root < 0)
                   root += nrou;
#endif
                 if(momentum_prefac[dind] != momentum_prefac[i])
                    prefac *= sqrt(((REAL) momentum_prefac[i]) /
                                   ((REAL) momentum_prefac[dind]));

#ifndef KNZ
                 dstdd.elements[dind].re += (prefac * src->elements[i].re);
                 dstdd.elements[dind].im += (prefac * src->elements[i].im);
#else
                 multiply(src->elements+i, rou_val+root, &tempc);
                 dstdd.elements[dind].re += (prefac * tempc.re);
                 dstdd.elements[dind].im += (prefac * tempc.im);
#endif
                }
              }
             state.down = tempd;
           }
#ifndef KNZ
           }
#endif
        }
      }
     sprod(src, &dstuu, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "SCuu(%d) := ", x+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");

     sprod(src, &dstdd, &ctmpb);
     divide(&ctmpb, &ctmpa, &ctmpc);

     fprintf(fpo, "SCdd(%d) := ", x+1);
     output_complex(&ctmpc, fpo);
     fprintf(fpo, ";\n");
   }

  free_vector(&dstdd);
  free_vector(&dstud);
  free_vector(&dstuu);
 }

/* Read parameters */

void read_params()
 {
  FILE *fpi;
  char temp_str[256];
  register int i;
  int line;

  fpi  = fopen("internal-params", "r");
  if(! fpi)
   {
    fprintf(stderr, "### Error: Failed to open `internal-params' for reading\n");
    exit(-14);
   }
  line = 0;
  while(fgets(temp_str, 255, fpi))
   {
    line ++;
    for(i=0; temp_str[i] && (temp_str[i] <= ' ') && (i < 255); i++)
          ;
    if(temp_str[i] && (temp_str[i] != '#') && (temp_str[i] != '%'))
     {
      if(temp_str[i+1] != '=')
        {
         fprintf(stderr, "### Error: Not a `=' as second character on line %d in file `internal-params'\n", line);
         exit(-14);
         }
       switch(temp_str[i])
         {
           case 'l':
             sscanf(temp_str+i+2, "%d", &size);
             break;
           case 'u':
             sscanf(temp_str+i+2, "%d", &n_part_u);
             break;
           case 'd':
             sscanf(temp_str+i+2, "%d", &n_part_d);
             break;
           case 'P':
             sscanf(temp_str+i+2, "%d", &period);
             break;
           case 'k':
#ifndef KNZ
             momentum = (temp_str[i+2] == '-')?-1:1;
#else
             sscanf(temp_str+i+2, "%d", &momentum);
#endif
             break;
#ifndef KNZ
           case 'p':
             parity = (temp_str[i+2] == '-')?-1:1;
             break;
#endif
           case 's':
             spin_inversion = (temp_str[i+2] == '-')?-1:1;
             break;
           case 't':
             sscanf(temp_str+i+2, "%lg", &tval);
             break;
           case 'T':
             sscanf(temp_str+i+2, "%lg", &tpval);
             break;
           case 'U':
             sscanf(temp_str+i+2, "%lg", &Uval);
             break;
           default:
             temp_str[i+1] = 0;
             fprintf(stderr, "### Error: Unknown `%s' as first character on line %d in file `internal-params'\n", temp_str+i, line);
             exit(-14);
         }
     }
   }
  fclose(fpi);

  if((size <= 0) || (size > MAXD))
   {
    fprintf(stderr, "### Error: size is not between 0 and %d\n", MAXD);
    exit(-14);
   }

  if(size % period)
   {
    fprintf(stderr, "### Error: period %d does not divide size %d\n", period, size);
    exit(-14);
   }

  if((n_part_u < 0) || (n_part_u > size))
   {
    fprintf(stderr, "### Error: Number of up electrons is not between 0 and %d\n", size);
    exit(-14);
   }

  if((n_part_d < 0) || (n_part_d > size))
   {
    fprintf(stderr, "### Error: Number of down electrons is not between 0 and %d\n", size);
    exit(-14);
   }

#ifdef KNZ
  if((momentum < 0) || (momentum > size/period))
   {
    fprintf(stderr, "### Error: Momentum is not between 0 and %d\n", size/period);
    exit(-17);
   }

  fprintf(stderr, "\nThis version does not support reflection symmetry\n");
#else
  fprintf(stderr, "\nThis version supports only momenta with exp(2 Pi I k/%d) = +/- 1\n", size/period);
#endif

  fprintf(stderr, "\nsize: %d\nup electrons: %d\ndown electrons: %d\n",
         size, n_part_u, n_part_d);
  fprintf(stderr, "period: %d\n", period);

  fprintf(stderr, "t = %1.12lg\n", tval);
  fprintf(stderr, "t' = %1.12lg\n", tpval);
  fprintf(stderr, "U = %1.12lg\n", Uval);
#ifdef KNZ
  fprintf(stderr, "momentum: %d\n", momentum);
#else
  fprintf(stderr, "momentum: %s\n", (momentum == -1)?"-":"+");
  if(parity)
    fprintf(stderr, "parity: %s\n", (parity == -1)?"-":"+");
  else
    fprintf(stderr, "no parity\n");
#endif
  if(n_part_u == n_part_d)
    fprintf(stderr, "Spin inversion: %s\n", (spin_inversion == -1)?"-":"+");
 }

/* Init the internal Hamiltonian application */

long init_interal_h()
 {
  register int i;

  fprintf(stderr, "Internal application of the Hamilton-Matrix\n");
  fprintf(stderr, "for the 1D p-merized Hubbard model\n");
  fprintf(stderr, "written December 1998-March 2003 by Andreas Honecker\n");

  read_params();

  init_mtables();

  gen_states();

  fprintf(stderr, "\nNumber of states with %d up and %d down electrons on a length %d chain: %ld\n",
                   n_part_u, n_part_d, size, nstates); 

  if(nstates >= MAXNSTATES)
   {
    fprintf(stderr, "### Error: This exceeds the storage capacity of the LONG data type\n");
    exit(113);
   }

  create_momenta();

  fprintf(stderr, "\nNumber of associated momentum eigenstates: %ld\n",
                   pstates);

  return(pstates);
 }

/* Exit the internal Hamiltonian application */

void exit_interal_h()
 {
  free(momentum_prefac);		/* Now free the tables used */
  free(momentum_charge);
#ifdef KNZ
  free(rou);
#else
  free(charge_prefac);
#endif
  free(charge_momentum);
#ifdef KNZ
  free(rou_val);
#endif
  
  free_states();

  free_mtables();
 }
