/* *** February 8, 2003: Correlation functions added
   *** June 17-19, 2004: Some data types changed to save memory
   *** August 18, 2004: Made XXZ anisotropy DELTA variable
       (replace the "double DELTA = 1.0L;" by "#define DELTA 1.0L"
        to switch back to a precompiled Delta = 1)
   *** April 17-20, 2005: Some data types modified to allow chains
       with up to 64 spins
   *** April 20, 2005: Added correlators of z-component of
       chirality: kappa^z_x = i/2 (S^+_x S^-_{x+1} - S^+_x S^+_{x+1})
*/

/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "matrix.h"
#include "vectors.h"
#include "apply_h.h"

/* data type for spin representation -
   this defines the available linear extent */

/* typedef unsigned long LLONG; */
typedef unsigned long long LLONG;

/* The parameters */

/* #define DELTA 1.0L */
double DELTA = 1.0L;
double jval = 1.0L;
double jpval = 1.0L;

/* The maximal number of states we can handle */

#define MAXNSTATES ( ((unsigned long) 1) << (sizeof(LONG)*8 - 1) ) + \
     ( ( ((unsigned long) 1) << (sizeof(LONG)*8 - 1) ) - 1000)

/* ... and the momentum-subspaces */

#define MAXSNSTATES ( ((unsigned long) 1) << (sizeof(SLONG)*8 - 1) ) + \
     ( ( ((unsigned long) 1) << (sizeof(SLONG)*8 - 1) ) - 1000)

/* Two special numbers (should be negative and different, otherwise arbitrary) */

#define MAGIC (MAXSNSTATES+123)
#define MAGIX (MAXSNSTATES+876)

/* The maximal linear extent */

#define MAXD (8*sizeof(LLONG))

/* The linear size */

static int size=MAGIC;

/* The number of particles (corresponding to total S^z) */

static int n_part=MAGIC;

/* The various parities */

static int momentum=1;		/* momentum */
static int parity=1;		/* reflection */

/* The number of states in the total-S^z vector space */

static unsigned long nstates;

/* Dimension of the momentum eigenspace */

static long pstates;

/* Mapping tables */

static SLONG *charge_momentum;
	/* Which S^z eigenvector belongs to which momentum eigenvector ? */

static signed char *charge_prefac;
	/* The (sign) prefactors */

static LONG *momentum_charge;
	/* Representative S^z eigenvector for the momentum eigenvector */

static INT *momentum_prefac;
	/* The number of states associated to the momentum eigenvector */

/* Mapping tables */

#ifndef SLOWINDEX
LONG ***index_tab;
char *np_tab;

int tab_size;

/* The maximal number of bits used in these tables */

static long maxbits;

/* The corresponding mask */

static unsigned long mapmask;

/* In how many slices do we split a state for the mapping  ? */

static int map_split;

#endif

/* Table of binomial coefficients */

static long *bin_coeff[1200];

/* ****** Some math routines ****** */

/* define DIRECTBIN in order not to store, but to compute the binomial
   coefficients - useful for very large systems at small densities  */

#ifdef DIRECTBIN
  #define over(a,b) binomial(a,b)
#else
  #define over(a,b) bin_coeff[a][b]
#endif

/* Den groessten gemeinsamen Teiler zweier Zahlen feststellen
   ueber Euklidischen Algorithmus
   Zehnerpotenzen schneller als Primfaktorzerlegung */

static long long get_gcd(long long a, long long b)
 {
   register int i;
   long long remainder;

   if(a == 0)           /* Sonderbehandlung fuer Nullen */
     return(1);
   if(b == 0)
     return(1);

   if(a < 0)            /* Betrag bilden */
     a=-a;
   if(b < 0)
     b=-b;

   if(a == 1)           /* Sonderbehandlung fuer Einsen */
     return(1);
   if(b == 1)
     return(1);

   remainder = a % b;   /* Das ist der Euklische Algorithmus */
   while(remainder)
    {
      a=b;
      b=remainder;
      remainder = a % b;
    }           /* Nun steht der GGT in b */

   return(b);
 }

/* Der Binomialkoeffizient */

static long binomial(long n, long m)
 {
   register long i;
   register long long gcd;
   register long long res1, res2;

   if(m > n-m)                  /* Symmetrie ausnutzen */
     m = n-m;

   if(m < 0)
     return(0);

#ifdef PRECOMP
   if((n==30) && (m==15))
     return(155117520L);

   if((n==31) && (m==13))
     return(206253075L);
   if((n==31) && (m==14))
     return(265182525L);
   if((n==31) && (m==15))
     return(300540195L);

   if((n==32) && (m==12))
     return(225792840L);
   if((n==32) && (m==13))
     return(347373600L);
   if((n==32) && (m==14))
     return(471435600L);
   if((n==32) && (m==15))
     return(565722720L);
   if((n==32) && (m==16))
     return(601080390L);
#endif

   res1 = 1;
   res2 = 1;
   for(i=0; i<m; i++)
    {
      res1 *= (n-i);           /* Multiplizieren */
      res2 *= (i+1);
      gcd = get_gcd(res1, res2);      /* Und kuerzen */
      res1 /= gcd;
      res2 /= gcd;
    }
   if(res2 != 1)
     fprintf(stderr, "*** Error: Binomial coefficient (%ld,%ld) not an integer !\n", n, m);
   return(res1);
 }

/* Allocate memory and produce error message if not available */

static void *m_alloc(long size, char *str)
 {
  void *ptr;

  ptr = (void *)malloc(size);

  if(! ptr)
   {
    fprintf(stderr, "### Error: Out of memory in %s()\n", str);
    exit(-13);
   }
  return(ptr);
 }

/* Initialize binomial coefficients */

static void init_binomials()
 {
  register long n, m;

  for(n=0; n<=size; n++)
   {
    bin_coeff[n] = (long *)m_alloc((((n+1)<n_part)?(n+2):(n_part+1))*sizeof(long), "init_binomials");
    for(m=0; m<=n+1; m++)
     if(m <= n_part)
       bin_coeff[n][m] = binomial(n,m);
   }
 }

/* Free binomial coefficients */

static void free_binomials()
 {
  register long n;

  for(n=0; n<=size; n++)
    free(bin_coeff[n]);
 }

#ifndef SLOWINDEX

static char IMSTR[] = "init_mtables";

/* Initialize mapping tables */

static void init_mtables()
 {
  register int i, j, x;
  long mask;
  LONG index;
  int cnt, np, lnp;

  /* to split into pieces of at most 16 bits
     such that splitting divides system size */
  map_split = 1;
  while((size/map_split > 16) && (! (size % (map_split+map_split))))
         map_split *= 2;
  while((size/map_split > 16) && (! (size % (map_split+map_split+map_split))))
         map_split *= 3;
  maxbits = size/map_split;
  mapmask = (1<<maxbits)-1;
#ifdef DEBUG
  printf("map_split = %d\n", map_split);
  printf("maxbits = %d\n", maxbits);
  printf("mapmask = %lx\n", mapmask);
#endif

  tab_size = 1<<maxbits;

  np_tab = (char *)m_alloc(tab_size * sizeof(char), IMSTR);

  for(i=0; i<tab_size; i++)
   {
    cnt = 0;
    mask = 1;
    for(x = 0; x<maxbits; x++)
     {
       if(i & mask)
         cnt++;
       mask = mask<<1;
     }
    np = np_tab[i] = cnt;
   }

  index_tab = (LONG ***)m_alloc(tab_size * sizeof(LONG **), IMSTR);
  for(i=0; i<tab_size; i++)
   {
    index_tab[i] = (LONG **)m_alloc(map_split * sizeof(LONG *), IMSTR);
    for(j=0; j<map_split; j++)
     {
      if(n_part - np_tab[i] >= 0)
        index_tab[i][j] = (LONG *)m_alloc((n_part - np_tab[i] + 1)
                  * sizeof(LONG), IMSTR);
      else
        index_tab[i][j] = 0;
     }
   }

  for(i=0; i<tab_size; i++)
   for(j=0; j<map_split; j++)
    for(lnp=0; lnp<=n_part - np_tab[i]; lnp++)
     {
      np = lnp + np_tab[i];
      cnt = j*maxbits;
      index = 0;
      mask = ((LLONG) 1)<<(maxbits-1);
      for(x=0; x<maxbits; x++)
       {
        if(i & mask)
         {
          index += over(size-cnt-1, np);
          np--;
          if(np < 0)
           fprintf(stderr, "### Error: Illegal bit pattern %x encountered:\n", i);
         }
        cnt++;
        mask = mask>>1;
       }
      index_tab[i][j][lnp] = index; 
     }
 }

/* Free mapping tables */

static void free_mtables()
 {
  register int i, j;

  for(i=0; i<tab_size; i++)
   {
    for(j=0; j<map_split; j++)
       if(index_tab[i][j])
           free(index_tab[i][j]);
    free(index_tab[i]);
   }
  free(index_tab);
  free(np_tab);
 }
#endif

/* Show a state */

static void show_state(FILE *out, LLONG map)
 {
  register LLONG mask;
  register int i;

  fputc('|', out);
  mask = (((LLONG) 1)<<(size-1));
  for(i=0; i<size; i++)
   {
    fputc((map&mask)?'1':'0', out);
    mask = mask>>1;
   }
  fputc('>', out);
  fputc('\n', out);
 }


/* Generate states in S^z-symmetry subspace */

static void gen_states()
 {
  nstates = over(size, n_part);
 }

/* Free states in S^z-symmetry subspace */

static void free_states()
 {
 }

/* Convert index to bit representation of state */

static LLONG index_to_state(LONG index)
 {
  LONG oindex;
  LLONG state;
  LLONG mask;
  int np, hh;
  int count;
  register int i, j, k;

  state = 0;

  oindex = index;

  count = 0;
  np = n_part;

  mask = ((LLONG) 1)<<(size-1);
  for(i=0; i<size; i++)
   {
    hh = over(size-count-1, np);
#ifdef DEBUG
    fprintf(stderr, "(%d,%d) = %d\n", size-count-1, np, hh);
#endif
    if(index >= hh)
     {
      index -= hh;
      np--;
      state |= mask;
     }
    count++;
    mask = mask>>1;
   }

  if(index || (np < 0))
   fprintf(stderr, "*** Error: failed to translate index `%ld' to state\n", oindex);

#ifdef DEBUG
  fprintf(stderr, "Result of index_to_state: %lx\n", state);
#endif

  return(state);
 }

/* Convert bit representation of state to index of matrix */

static LONG state_to_index(LLONG state)
 {
  LONG index;
#ifdef SLOWINDEX
  LLONG mask;
  int count;
#else
  LLONG cstate;
  int lnp;
#endif
  int np;
  register int i, j, k;

#ifdef SLOWINDEX
  count = index = 0;
  np = n_part;
  mask = ((LLONG) 1)<<(size-1);
  for(i=0; i<size; i++)
   {
    if(state & mask)
     {
      index += over(size-count-1, np);
      np--;
      if(np < 0)
       {
        fprintf(stderr, "### Error: Illegal state encountered: ");
        show_state(stderr, state);
       }
     }
    count++;
    mask = mask>>1;
   }
#else
  index = 0;
  np = n_part;
  for(j=0; j<map_split; j++)
   {
    if(j==(map_split-1))
      cstate = state & mapmask;
    else
      cstate = (state>>(maxbits*(map_split-1-j))) & mapmask;
    lnp = np-np_tab[cstate];
    if(lnp < 0)
     {
      fprintf(stderr, "### Error: Illegal state encountered: ");
      show_state(stderr, state);
     }
    else
      index += index_tab[cstate][j][lnp];
    np = lnp;
   }
#endif

#ifdef DEBUG
  fprintf(stderr, "State: %lx\n", state);
  fprintf(stderr, "Index: %ld\n", index);
#endif

  return(index);
 }

/* Translate a state bitwise left */

static LLONG translate_left(LLONG state)
 {
  register LLONG mask;

  mask = ((LLONG) (((LLONG) 1)<<size)) - 1;
  if(! mask)
    mask = -1;
  state = (((state<<1) | ((state >> (size - 1)) & 1)) & mask);
  return(state);
 }

/* Reflect state */

static LLONG reflect(LLONG state)
 {
  register int j, x;
  LLONG temp;
  LLONG mask, rmask;

  temp = 0;
  rmask = mask = 1;
  for(x = 0; x<size; x++)
   {
    if(state & mask)
      temp |= rmask;
    mask = (mask << 1);
    if(rmask == 1)
      rmask = ((LLONG) 1) << (size-1);
    else
      rmask = (rmask >> 1);
   }
  return(temp);
 }

/* Create the momentum basis */

static void create_momenta()
 {
  register long i;
  register int j, k;
  int flag;
  int tcnt;
  LLONG state;
  int refl;
  int sign, fac;
  LONG index;
  LONG ps[3000];
  int pcnt;

  charge_momentum = (SLONG *)m_alloc(sizeof(SLONG)*nstates, "create_momenta");
  charge_prefac = (signed char *)m_alloc(sizeof(signed char)*nstates, "create_momenta");

  for(i=0; i<nstates; i++)
   {
    charge_momentum[i] = MAGIC;
    charge_prefac[i] = 0;
   }

  pstates = 0;
  for(i=0; i<nstates; i++)
   if(charge_momentum[i] == MAGIC)
    {
     state = index_to_state(i);
     sign = 1;
     pcnt = 0;

     for(refl=0; refl<=1; refl++)
      {				/* Orbit over reflection group */
       for(j=0; j<size; j++)
        {			/* Orbit over translational group */
         index = state_to_index(state);
         charge_momentum[index] = pstates;   
         charge_prefac[index] += sign;

         for(k=0; (k<pcnt) && (ps[k] != index); k++)
                  ;		/* Remember all states which appeared */
         if(k==pcnt)
            ps[pcnt++] = index;

         state=translate_left(state);
         sign *= momentum;
        }			/* End of translation group */
       state = reflect(state);
       sign *= parity;
      }				/* End of reflection group */

     fac = -1;			/* Test prefactors */
     for(j=0; j<pcnt; j++)
        {
         sign = charge_prefac[ps[j]];
         if(sign < 0)
           sign = -sign;
         if(sign)
           charge_prefac[ps[j]] /= sign;
         if(fac < 0)
           fac = sign;
         if(fac != sign)
          {
           fprintf(stderr, "### Error: Not all prefactors have equal magnitude\n");
           exit(-15);
          }
        }

     if(! fac)
      {
       for(j=0; j<pcnt; j++)
           charge_momentum[ps[j]] = MAGIX;
      }
     else
       pstates++;
    }

  momentum_charge = (LONG *)m_alloc(sizeof(LONG)*pstates, "create_momenta");
  momentum_prefac = (INT *)m_alloc(sizeof(INT)*pstates, "create_momenta");

  for(i=0; i < pstates; i++)
   {
    momentum_charge[i] = MAGIC;
    momentum_prefac[i] = 0; 
   }

  for(i=0; i<nstates; i++)
   if(charge_momentum[i] < MAXSNSTATES)
     {
       if(momentum_charge[charge_momentum[i]] >= MAXSNSTATES)
         momentum_charge[charge_momentum[i]] = i;
       momentum_prefac[charge_momentum[i]]++;
     }
 }

/* Apply Hamilton operator */

void apply_internal_ham(struct vector *src, struct vector *dst, char *name)
 {
  register long i, j;
  register SLONG sind;
  register int x;
  int nj;
  LLONG mask, nmask, nnmask;
  LLONG tempa, tempb;
  LONG index;
  int szsz, szszp;
  LLONG state;
  REAL prefac;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  if(dst->el_cnt != pstates)
   {
    display_error("Dimension of output vector doesn't match - can't do internal matrix application");
    return;
   }

  for(i=0; i<dst->el_cnt; i++)		/* First set destination to zero */
   {
     dst->elements[i].im = dst->elements[i].re = 0.0;
   }

  for(i=0; i<pstates; i++)		/* Loop over all momentum states */
   {
    szszp = szsz = 0;
    state = index_to_state(momentum_charge[i]);

    mask = 1;
    for(x=0; x<size; x++)		/* Loop over all sites */
     {
      nmask = mask<<1;			/* Nearest neighbour interaction */
      if(x == (size-1))
        nmask = 1;
      tempa = state;
      if( ((tempa & mask) && (! (tempa & nmask))) ||
          ((! (tempa & mask)) && (tempa & nmask)) )
       {				/* S^+ S^- - interaction */
         state ^= (mask | nmask);
         sind = charge_momentum[index = state_to_index(state)];
         if(sind < MAXSNSTATES)
          {
           prefac = (jpval/2.0L);
           if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
              prefac = -prefac;
           if(momentum_prefac[sind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[sind]));
           dst->elements[i].re += (prefac * src->elements[sind].re);
           dst->elements[i].im += (prefac * src->elements[sind].im);
          }
         state = tempa;
         szszp--;
       }
      else
         szszp++;

      nnmask = nmask<<1;		/* Next nearest neighbour interaction */
      if(x == (size-2))
        nnmask = 1;
      tempa = state;
      if( ((tempa & mask) && (! (tempa & nnmask))) ||
          ((! (tempa & mask)) && (tempa & nnmask)) )
       {                                /* S^+ S^- - interaction */
         state ^= (mask | nnmask);
         sind = charge_momentum[index = state_to_index(state)];
         if(sind < MAXSNSTATES)
          {
           prefac = (jval/2.0L);
           if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
              prefac = -prefac;
           if(momentum_prefac[sind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[sind]));
           dst->elements[i].re += (prefac * src->elements[sind].re);
           dst->elements[i].im += (prefac * src->elements[sind].im);
          }
         state = tempa;
         szsz--;
       }
      else
         szsz++;
        
      mask = nmask;
     }

    if(DELTA)
     {
      prefac = DELTA/4.0L*(jval*((REAL) szsz) + jpval*((REAL) szszp));	/* The diagonal element */
      dst->elements[i].re += (prefac * src->elements[i].re);
      dst->elements[i].im += (prefac * src->elements[i].im);
     }
   }
 }

/* Routine to compute correlation functions with respect to the vector in src
   (put an empty one, if you don't need it) */

void compute_correl(struct vector *src, FILE *fpo)
 {
  COMPLEX ctmpa, ctmpb, ctmpc;
  struct vector dst, dstpm;
  int x, x0, nx; 
  int x1, nx1;
  register long i;
  register SLONG sind;
  LLONG mask, nmask, mask1, nmask1;
  LLONG tempa;
  LONG index;
  int szsz;
  LLONG state;
  REAL prefac;
  COMPLEX resultpm;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  alloc_int_vector(&dst, pstates);
  alloc_int_vector(&dstpm, pstates);

  sprod(src, src, &ctmpa);

  fprintf(fpo, "\nvector Szz(%d);\n\n", size);
  fprintf(fpo, "vector Spm(%d);\n\n", size);

  for(x = 0; x<size; x++)
   {
    for(i=0; i<pstates; i++)		/* Loop over all momentum states */
     {
      szsz = 0;
      state = index_to_state(momentum_charge[i]);
      resultpm.re = resultpm.im = 0;

      for(x0=0; x0<size; x0++)		/* Loop over all sites */
       {
        mask = ((LLONG) 1)<<x0;
        nx = x0+x;
        while(nx >= size)
          nx -= size;
        nmask = ((LLONG) 1)<<nx;
        tempa = state;
        if( ((tempa & mask) && (! (tempa & nmask))) ||
            ((! (tempa & mask)) && (tempa & nmask)) )
         {				/* S^+ S^- - correlator */
           state ^= (mask | nmask);
           sind = charge_momentum[index = state_to_index(state)];
           if(sind < MAXSNSTATES)
            {
             prefac = (((REAL) 1)/((REAL) (2*size)));
             if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
                prefac = -prefac;
             if(momentum_prefac[sind] != momentum_prefac[i])
              {
                prefac *= sqrt(((REAL) momentum_prefac[i]) /
                               ((REAL) momentum_prefac[sind]));
              }
             resultpm.re += (prefac * src->elements[sind].re);
             resultpm.im += (prefac * src->elements[sind].im);
            }
           state = tempa;
           szsz--;
         }
        else
         {
          if(nx == x0)
           {
             prefac = (((REAL) 1)/((REAL) (2*size)));
             resultpm.re += (prefac * src->elements[i].re);
             resultpm.im += (prefac * src->elements[i].im);
           }
          szsz++;
         }
       }
      dstpm.elements[i].re = resultpm.re;
      dstpm.elements[i].im = resultpm.im;
      prefac = ((REAL) szsz)/((REAL) (4*size));
      dst.elements[i].re = (prefac * src->elements[i].re);
      dst.elements[i].im = (prefac * src->elements[i].im);
     }
    sprod(src, &dst, &ctmpb);
    divide(&ctmpb, &ctmpa, &ctmpc);
 
    fprintf(fpo, "Szz(%d) := ", x+1);
    output_complex(&ctmpc, fpo);
    fprintf(fpo, ";\n");
 
    sprod(src, &dstpm, &ctmpb);
    divide(&ctmpb, &ctmpa, &ctmpc);
 
    fprintf(fpo, "Spm(%d) := ", x+1);
    output_complex(&ctmpc, fpo);
    fprintf(fpo, ";\n");
   }

  free_vector(&dstpm);

  fprintf(fpo, "\nvector Kappazz(%d);\n\n", size);

  for(x = 2; x<size-1; x++)
   {
    for(i=0; i<pstates; i++)		/* Loop over all momentum states */
     {
      state = index_to_state(momentum_charge[i]);
      resultpm.re = resultpm.im = 0;

      for(x0=0; x0<size; x0++)		/* Loop over all sites */
       {
        mask = ((LLONG) 1)<<x0;
        x1 = x0+1; 
        nx = x0+x;
	nx1 = nx+1;
        while(x1 >= size)
          x1 -= size;
        while(nx >= size)
          nx -= size;
        while(nx1 >= size)
          nx1 -= size;
        mask1 = ((LLONG) 1)<<x1;
        nmask = ((LLONG) 1)<<nx;
        nmask1 = ((LLONG) 1)<<nx1;
        tempa = state;
        /* First kappaz-Operator */
        if( ((state & mask) && (! (state & mask1))) ||
            ((! (state & mask)) && (state & mask1)) )
         {
           prefac = (((REAL) -1)/((REAL) (4*size)));
	   if(state & mask)
              prefac = -prefac;
           state ^= (mask | mask1);
	 }
	else
	 prefac = 0;
        /* Second kappaz-Operator */
        if( ((state & nmask) && (! (state & nmask1))) ||
            ((! (state & nmask)) && (state & nmask1)) )
         {
	   if(state & nmask)
              prefac = -prefac;
           state ^= (nmask | nmask1);
	 }
	else
	 prefac = 0;
        if(prefac)
	 {
           sind = charge_momentum[index = state_to_index(state)];
           if(sind < MAXSNSTATES)
            {
             if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
                prefac = -prefac;
             if(momentum_prefac[sind] != momentum_prefac[i])
              {
                prefac *= sqrt(((REAL) momentum_prefac[i]) /
                               ((REAL) momentum_prefac[sind]));
              }
             resultpm.re += (prefac * src->elements[sind].re);
             resultpm.im += (prefac * src->elements[sind].im);
            }
         }
        state = tempa;
       }
      dst.elements[i].re = resultpm.re;
      dst.elements[i].im = resultpm.im;
     }
    sprod(src, &dst, &ctmpb);
    divide(&ctmpb, &ctmpa, &ctmpc);
 
    fprintf(fpo, "Kappazz(%d) := ", x+1);
    output_complex(&ctmpc, fpo);
    fprintf(fpo, ";\n");
   }

  free_vector(&dst);
 }

/* Read parameters */

void read_params()
 {
  FILE *fpi;
  char temp_str[256];
  register int i;
  int line;

  fpi  = fopen("internal-params", "r");
  if(! fpi)
   {
    fprintf(stderr, "### Error: Failed to open `internal-params' for reading\n");
    exit(-14);
   }
  line = 0;
  while(fgets(temp_str, 255, fpi))
   {
    line ++;
    for(i=0; temp_str[i] && (temp_str[i] <= ' ') && (i < 255); i++)
          ;
    if(temp_str[i] && (temp_str[i] != '#') && (temp_str[i] != '%'))
     {
      if(temp_str[i+1] != '=')
        {
         fprintf(stderr, "### Error: Not a `=' as second character on line %d in file `internal-params'\n", line);
         exit(-14);
         }
       switch(temp_str[i] | 0x20)
         {
           case 'l':
             sscanf(temp_str+i+2, "%d", &size);
             break;
           case 'n':
             sscanf(temp_str+i+2, "%d", &n_part);
             break;
           case 'k':
             momentum = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'p':
             parity = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'j':
             sscanf(temp_str+i+2, "%lg", &jval);
             break;
           case 'd':
#ifdef DELTA
             fprintf(stderr, "### Error: A value Delta=1 has been precompiled\n");
#else
             sscanf(temp_str+i+2, "%lg", &DELTA);
#endif
             break;
           case 'x':
             sscanf(temp_str+i+2, "%lg", &jpval);
             break;
           default:
             temp_str[i+1] = 0;
             fprintf(stderr, "### Error: Unknown `%s' as first character on line %d in file `internal-params'\n", temp_str+i, line);
             exit(-14);
         }
     }
   }
  fclose(fpi);

  if((size <= 0) || (size > MAXD))
   {
    fprintf(stderr, "### Error: size is not between 0 and %d\n", MAXD);
    exit(-14);
   }

  if((n_part < 0) || (n_part > size))
   {
    fprintf(stderr, "### Error: Number of particles is not between 0 and %d\n", size);
    exit(-14);
   }

  fprintf(stderr, "\nsize: %d\nnumber of particles: %d\n", size, n_part);

  fprintf(stderr, "J = %1.12lg\n", jval);
  fprintf(stderr, "J' = %1.12lg\n", jpval);
#ifndef DELTA
  fprintf(stderr, "Delta = %1.12lg\n", DELTA);
#endif
  fprintf(stderr, "momentum: %s\n", (momentum == -1)?"-":"+");
  fprintf(stderr, "parity: %s\n", (parity == -1)?"-":"+");
 }

/* Init the internal Hamiltonian application */

long init_interal_h()
 {
  register int i;

  fprintf(stderr, "Internal application of the Hamilton-Matrix\n");
  fprintf(stderr, "for the frustrated 1D spin-1/2 Heisenberg model\n");
  fprintf(stderr, "written December 1998-April 2005 by Andreas Honecker\n");

  read_params();

#ifndef DIRECTBIN
  init_binomials();
#endif
#ifndef SLOWINDEX
  init_mtables();
#endif

  gen_states();

  fprintf(stderr, "\nNumber of states with %d down spins on a length %d chain: %ld\n",
                   n_part, size, nstates); 

  if(nstates >= MAXNSTATES)
   {
    fprintf(stderr, "### Error: This exceeds the storage capacity of the LONG data type\n");
    exit(113);
   }

  create_momenta();

  fprintf(stderr, "\nNumber of associated momentum eigenstates: %ld\n",
                   pstates);

  if(pstates >= MAXSNSTATES)
   {
    fprintf(stderr, "### Error: This exceeds the storage capacity of the SLONG data type\n");
    exit(113);
   }

  return(pstates);
 }

/* Exit the internal Hamiltonian application */

void exit_interal_h()
 {
#ifndef SLOWINDEX
  free_mtables();
#endif
#ifndef DIRECTBIN
  free_binomials();
#endif

  free(momentum_prefac);		/* Now free the tables used */
  free(momentum_charge);
  free(charge_momentum);
  
  free_states();
 }
