/* **********************************************
   November 13-18, 2004:
     some optimization in tridiag_ql_implicit()
     and householder()
   February 15, 2005:
     eliminated overflow-problem in tridiag_ql_implicit()
   ********************************************** */

/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for this specific package */
#include "useful.h"
#include "complex.h"
#include "matrix.h"
#include "vectors.h"
#include "fullmatrix.h"

/* Free space occupied by a matrix */

void free_full_matrix(struct full_matrix *m)
 {
  register int i;

  if(m->bflag)				/* Allocated as one block ? */
    free(m->entries[0]);
  else
    for(i=0; i<m->rows; i++)
      free(m->entries[i]);
  if(m->rows)
    free(m->entries);
  m->rows = m->columns = 0;
 }

/* Allocate a full matrix */

void alloc_full_matrix(struct full_matrix *m, int rows, int columns, char bflag)
 {
   register int i, j;

   if((rows < 1) || (columns < 1))
    {
      display_error("Illegal arguments in definition of full matrix");
      return;
    }

   m->rows = m->columns = 0;

   m->entries = (COMPLEX **)malloc(rows * sizeof(COMPLEX *));
   if(! m->entries)
     goto error_exit;

   if(bflag)				/* Allocate as one block ? */
    {
     m->entries[0] = (COMPLEX *)malign16(columns * rows * sizeof(COMPLEX));
     if(! m->entries[0])
       goto error_exit;
     for(i=1; i<rows; i++)
       m->entries[i] = m->entries[0]+i*columns;
    }
   else
    {
     for(i=0; i<rows; i++)
      {
       m->entries[i] = (COMPLEX *)malign16(columns * sizeof(COMPLEX));
       if(! m->entries[i])
        {
         for(j=0; j<i; j++)
           free(m->entries[j]);
         goto error_exit;
        }
      }
    }

   /* Initialize to 0 */

#pragma omp parallel for private(i, j) if(columns>1000)
   for(i=0; i<rows; i++)
    for(j=0; j<columns; j++)
        m->entries[i][j].re = m->entries[i][j].im = 0;

   m->bflag = bflag;
   m->rows = rows;
   m->columns = columns;
   return;
error_exit:
   display_error("Not enough memory in allocation of full matrix");
 }

/* Display a full matrix */

void show_full_matrix(struct full_matrix *m)
 {
  register int i, j;

  for(i=0; i<m->rows; i++)
    for(j=0; j<m->columns; j++)
      {
       printf("[%d,%d] = ", i+1, j+1);
       show_complex(m->entries[i]+j);
      }
 }

/* Write a full matrix in processable form */

void write_full_matrix(char *name, struct full_matrix *m, FILE *fpo)
 {
  register int i, j;

  fprintf(fpo, "matrix %s(%d,%d);\n", name, m->rows, m->columns);

  for(i=0; i<m->rows; i++)
    for(j=0; j<m->columns; j++)
      {
       fprintf(fpo, "%s(%d,%d) := ", name, i+1, j+1);
       output_complex(m->entries[i]+j, fpo);
       fprintf(fpo, ";\n");
      }
 }

/* Use Jacobi's method to diagonalize a Hermitian matrix
   inspired by Sec. 11.1 of "Numerical Recipes in C", W.H. Press, S.A.
   Teukolsky, W.T. Vetterling, B.P. Flannery, Cambridge University Press (1992)

   (we are a bit less ambitious as far as efficiency and accuracy is concerned)

   input matrix is m, returns the transform that diagonalizes things in v
   m is destroyed upon return - it is m -> v^(-1) m v, which should
   be diagonal 
*/

#define MAXSWEEP 150			/* The maximal number of sweeps */

void jacobi_diagonalize(struct full_matrix *m, struct full_matrix *v, REAL accuracy)
 {
  int sweep;
  register int i, j, k;
  REAL sm, tresh, g;
  COMPLEX c, s, ss, t, theta;
  COMPLEX t1, t2, t3, t4;

  if((m->rows < 1) || (m->columns < 1))
   {
     display_error("Jacobi method called with improperly defined matrix");
     return;
   }

  if(m->rows != m->columns)
   {
     display_error("Jacobi method called with non-square matrix");
     return;
   }

  if(v->rows)			/* Initialize v to the identity-matrix */
    free_full_matrix(v);
  alloc_full_matrix(v, m->rows, m->rows, FALSE);
  for(i=0; i<v->rows; i++)
       v->entries[i][i].re = 1.0L;

  for(sweep = 0; sweep<MAXSWEEP; sweep++)
   {					/* Start of a sweep */
    sm = 0;				/* Sum off-diagonal elements. */
    for(i=0; i<m->rows; i++)
     for(j=i+1; j<m->columns; j++)
       sm += complex_abs(m->entries[i]+j);

    if(sm <= accuracy)			/* Return if residual sum is small */
     {
#ifdef DEBUGJACOBI
  printf("Jacobi method successfully completed after %d sweeps\n", sweep);
#endif
      return;
     }
    if(sweep < 3)
      tresh = sm/(5.0L * m->rows * m->columns);	/* ... on the first three sweeps */
    else
      tresh = 0;				/* ... thereafter */

    for(i=0; i<m->rows; i++)		/* All off-diagonal elements */
     for(j=i+1; j<m->columns; j++)
      {
       g = 100.0L * complex_abs(m->entries[i]+j); 
       if((sweep > 3)
         && ((complex_abs(m->entries[i]+i) + g) == complex_abs(m->entries[i]+i))
         && ((complex_abs(m->entries[j]+j) + g) == complex_abs(m->entries[j]+j)))
        {
         m->entries[i][j].re = m->entries[i][j].im = 0;
        }
       else
        if(complex_abs(m->entries[i]+j) > tresh)
          {
					/* (modified) Equation (11.1.8)
					   (also our sign convention for theta is different) */
           sub(m->entries[i]+i, m->entries[j]+j, &t1);
           t2.re = m->entries[i][j].re;		/* complex conjugate of off-diagonal element */
           t2.im = -m->entries[i][j].im;
           divide(&t1, &t2, &theta);
           theta.re *= 0.5L;
           theta.im *= 0.5L;

	/* Instead of eq. (11.1.10) we use
		t = theta +/- sqrt(theta^2 + a[p][q]/a[p][q]^*)
	   and choose the smaller root */

           if(complex_abs_sq(&theta) < 1e18)
            {
             divide(m->entries[i]+j, &t2, &t1);
             multiply(&theta, &theta, &t2);
             add(&t1, &t2, &t3);
             complex_sqrt(&t3, &t1);
             add(&theta, &t1, &t);
             sub(&theta, &t1, &t2);
             if(complex_abs_sq(&t2) < complex_abs_sq(&t))
              {					/* take the smaller root */
               t.re = t2.re;
               t.im = t2.im;
              }
            }
           else
            {				/* If theta larger, expand the sqrt */
             divide(m->entries[i]+j, &t2, &t1);
             divide(&t1, &theta, &t);
             t.re *= -0.5L;
             t.im *= -0.5L;
            }
	/* We choose c real */
           c.re = 1.0L/sqrt(complex_abs_sq(&t) + ((double) 1));		/* (11.1.11) */
           c.im = 0;
           multiply(&t, &c, &s);				/* (11.1.12) */
           ss.re = s.re;					/* The complex conjugate thereof */
           ss.im = -s.im;
#ifdef DEBUGJACOBI
  printf("theta = ");
  show_complex(&theta);
  printf("t = ");
  show_complex(&t);
  printf("c = ");
  show_complex(&c);
  printf("s = ");
  show_complex(&s);
  printf("s^* = ");
  show_complex(&ss);
#endif
	/* Multiply matrices m and v from right with
		/        \
		| c    s |
		| -s^* c |
                \        /    */
           for(k=0; k<m->columns; k++)
            {
             multiply(m->entries[k]+i, &c, &t1);		/* first m */
             multiply(m->entries[k]+j, &ss, &t2); 
             multiply(m->entries[k]+i, &s, &t3);
             multiply(m->entries[k]+j, &c, &t4);
             sub(&t1, &t2, m->entries[k]+i);
             add(&t3, &t4, m->entries[k]+j);
             multiply(v->entries[k]+i, &c, &t1);		/* then v */
             multiply(v->entries[k]+j, &ss, &t2);
             multiply(v->entries[k]+i, &s, &t3);
             multiply(v->entries[k]+j, &c, &t4);
             sub(&t1, &t2, v->entries[k]+i);
             add(&t3, &t4, v->entries[k]+j);
            }
	/* Multiply matrix m from left with
		/        \
		| c   -s |
		| s^*  c |
		\        /    */
           for(k=0; k<m->rows; k++)
            {
             multiply(&c, m->entries[i]+k, &t1);
             multiply(&s, m->entries[j]+k, &t2);
             multiply(&ss, m->entries[i]+k, &t3);
             multiply(&c, m->entries[j]+k, &t4);
             sub(&t1, &t2, m->entries[i]+k);
             add(&t3, &t4, m->entries[j]+k);
            }
#ifdef DEBUGJACOBI
  printf("Matrix after transformation:\n");
  show_full_matrix(m);
#endif
	/* Set the entry we eliminated to zero */
           m->entries[i][j].re = m->entries[i][j].im = 0;
          }  
      }
   }
  display_error("Jacobi method terminated without converging");
 }


/* Compute an adjoint */

void full_adjoint(struct full_matrix *s, struct full_matrix *d)
 {
  register int i, j;

  if((s->rows < 1) || (s->columns < 1))
   {
     display_error("Adjoint called with improperly defined matrix");
     return;
   }

  if(d->rows)					/* Initialize d */
    free_full_matrix(d);
  alloc_full_matrix(d, s->columns, s->rows, s->bflag);

#pragma omp parallel for private(i,j) if(d->rows > 1000)
  for(i=0; i<d->rows; i++)
   for(j=0; j<d->columns; j++)
     complex_conjugate(s->entries[j]+i, d->entries[i]+j);
 }

/* Copy a sparse matrix to a full matrix */

void copy_matrix_to_full(struct matrix *s, struct full_matrix *d)
 {
  register int i, j;
  struct matrix_el *ptr;

  if((s->rows != d->rows) || (s->columns != d->columns))
   {
     display_error("Tried to copy matrices of incompatible sizes");
     return;
   }

#pragma omp parallel for private(i, j) if(d->rows>1000)
   for(i=0; i<d->rows; i++)
    for(j=0; j<d->columns; j++)
        d->entries[i][j].re = d->entries[i][j].im = 0;

#pragma omp parallel for private(i,j,ptr) if(s->rows>5000)
   for(i=0; i<s->rows; i++)
     for(ptr=s->matrix_rows[i]; ptr; ptr=ptr->next)
       {
         j = ptr->column;
         d->entries[i][j].re = ptr->value.re;
         d->entries[i][j].im = ptr->value.im;
       }
 }

/* Test if a real number is zero */

#define ZERO 1e-13

int is_zero(REAL num)
 {
  if((num < ZERO) && (num > - ZERO))
    return(TRUE);
  return(FALSE);
 }

/* Swap two rows of a full matrix */

static void swap_rows(struct full_matrix *s, int row1, int row2)
 {
  int i;
  REAL re, im;

  if((row1 < 0) || (row1 >= s->rows))
   {
     display_error("swap_rows: row1 out of range\n");
     return;
   }
  if((row2 < 0) || (row2 >= s->rows))
   {
     display_error("swap_rows: row2 out of range\n");
     return;
   }

  for(i=0; i<s->columns; i++)
   {
    re = s->entries[row1][i].re;
    im = s->entries[row1][i].im;
    s->entries[row1][i].re = s->entries[row2][i].re;
    s->entries[row1][i].im = s->entries[row2][i].im;
    s->entries[row2][i].re = re;
    s->entries[row2][i].im = im;
   }
 }

/* Swap two columns of a full matrix */
 
static void swap_columns(struct full_matrix *s, int col1, int col2)
 {
  int i;
  REAL re, im;
 
  if((col1 < 0) || (col1 >= s->columns))
   {
     display_error("swap_columns: col1 out of range\n");
     return;
   }
  if((col2 < 0) || (col2 >= s->columns))
   {
     display_error("swap_columns: col2 out of range\n");
     return;
   }
 
  for(i=0; i<s->columns; i++)
   {
    re = s->entries[i][col1].re;
    im = s->entries[i][col1].im;
    s->entries[i][col1].re = s->entries[i][col2].re;
    s->entries[i][col1].im = s->entries[i][col2].im;
    s->entries[i][col2].re = re;
    s->entries[i][col2].im = im;
   }
 }

/* Some auxiliary stuff for the routine tridiag_ql_implicit() */

static struct tqli_entry { int i; REAL c; REAL s; }
       *tqli_buf=0;			/* Buffer */

static int tqli_cnt=0;

#define TQLIBUFSIZE 32768 /* 8192 */

/* execute transformation in buffer */

void tqli_buf_flush(struct full_matrix *bchg, int dim)
 {
  register int j, i, k;
  register REAL re, im;
  register REAL c, s;
  register COMPLEX *bptr;
  register struct tqli_entry *tptr;

#pragma omp parallel for private(k,j,i,c,s, re, im, bptr, tptr) if(dim > 500)
  for(k=0; k<dim; k++)			/* form eigenvectors */
   {
    /* Move out as many pointer computations as possible
       from inner loop */
    bptr = bchg->entries[k];
    tptr = tqli_buf;
    for(j=0; j<tqli_cnt; j++)
     {
      i = tptr->i;
      c = tptr->c;
      s = tptr->s;
      re  = bptr[i+1].re;
      im  = bptr[i+1].im;
      bptr[i+1].re = s*bptr[i].re+c*re;
      bptr[i+1].im = s*bptr[i].im+c*im;
      bptr[i].re   = c*bptr[i].re-s*re;
      bptr[i].im   = c*bptr[i].im-s*im;
      tptr++;
     }
   }

  tqli_cnt = 0;
 }


/* The sparse matrix "inp" is the Hermitean tri-diagonal matrix
   to be diagonalized.

   If "bchg" is a non-zero pointer, then the transformation required
   for the diagonalization is applied to it, i.e. if you want bchg
   to return the change of basis, you should initialize it to the
   identity matrix

   If "res" is a non-zero pointer, it returns the eigenvalues in
   ascending order */

#ifdef DEBUGDIAG
extern void debug_out(char *s);
#endif

void tridiag_ql_implicit(struct matrix *inp, struct full_matrix *bchg,
                         struct vector *res)
 {
   register int i, l, m, iter, k;
   REAL *d, *e;
   register REAL s,r,p,g,f,dd,c,b;
   register REAL re, im;
   COMPLEX tmp1, tmp2;
   int dim;
   struct matrix_el *ptr;
   register COMPLEX *bptr;
   int flag;
   struct vector pfac;
   int *smap;
   int mnum, mnum1;
   int temp_int;

#ifdef DEBUGDIAG
   debug_out("tridiag_ql_implicit() started");
#endif

   if((inp->rows < 1) || (inp->columns < 1))
    {
      display_error("Tridiagonal matrix diagonalization called with improperly defined matrix");
      return;
    }

   if(inp->rows != inp->columns)
    {
      display_error("Tridiagonal matrix diagonalization called with non-square matrix");
      return;
    }

   dim = inp->rows;

   d = (REAL *)malloc(dim * sizeof(REAL));
   if(! d)
    {
     display_error("Not enough memory for diagonal matrix elements of tridiagonal matrix");
     return;
    }
   e = (REAL *)malloc((dim-1) * sizeof(REAL));
   if(! e)
    {
     free(d);
     display_error("Not enough memory for subdiagonal matrix elements of tridiagonal matrix");
     return;
    }

   if(bchg)
    {
     tqli_buf = (struct tqli_entry *) malloc(TQLIBUFSIZE * sizeof(struct tqli_entry));
     if(! tqli_buf)
      {
       free(e);
       free(d);
       display_error("Not enough memory for basis change buffer");
       return;
      }
     tqli_cnt = 0;
    }

   alloc_int_vector(&pfac, dim);	/* vector for phase factors */
   for(l=0; l<dim; l++)
    {
     pfac.elements[l].re = 1;
     pfac.elements[l].im = 0;
    }

   /* Now we copy the input matrix into the vectors d and e
      - using appropriate phase factors to make them real */

   for(l=0; l<dim-1; l++)
      e[l] = 0;

   for(l=0; l<dim; l++)
    {
     d[l] = 0;
     for(ptr=inp->matrix_rows[l]; ptr; ptr=ptr->next)
      {
       switch(ptr->column - l)
        {
        case 0:
          d[l] = ptr->value.re;
          if(! is_zero(ptr->value.im/(d[l]?d[l]:1.0L)))
            display_error("TQLI: Matrix passed with diagonal element which is not real");
          break;
         case 1:
          e[l] = complex_abs(&(ptr->value));
          if(e[l])
           {
            pfac.elements[l].re = ptr->value.re/e[l];
            pfac.elements[l].im = ptr->value.im/e[l];
           }
          break;
         case -1:
          f = complex_abs(&(ptr->value));
          if(! is_zero((f-e[l-1])/(f?f:1.0L)))
            display_error("TQLI: Non-hermitean matrix passed");
          if(f)
           {
             if((! is_zero(ptr->value.re/f - pfac.elements[l-1].re))
              || (! is_zero(ptr->value.im/f + pfac.elements[l-1].im)))
                display_error("TQLI: Non-hermitean matrix passed");
           }
          break;
         default:
          display_error("TQLI: non-tridiagonal matrix passed");
          break;
        }
      }
    }

   /* now compute RELATIVE phase factors */

   for(l=dim-2; l>=0; l--)
    {
     multiply(pfac.elements+l, pfac.elements+l+1, &tmp1);
     pfac.elements[l].re = tmp1.re;
     pfac.elements[l].im = tmp1.im;
    }

   /* apply phase factors to the original input matrix
      (multiplication from the right) */

   if(bchg)
    {
#pragma omp parallel for private(i,k,tmp1,tmp2) if(dim>100)
     for(i=0; i<dim; i++)
      {
       tmp1.re = pfac.elements[i].re;
       tmp1.im = pfac.elements[i].im;
       /* apply this phase to ith column */
       for(k=0; k<dim; k++)
        {
         multiply(&tmp1, bchg->entries[k]+i, &tmp2);
         bchg->entries[k][i].re = tmp2.re;
         bchg->entries[k][i].im = tmp2.im;
        }
      }
    }
   free_vector(&pfac);

/* Now we use the QL algorithm with implicit shifts to determine the
   eigenvalues and vectors of a real, symmetric, tridiagonal matrix.

   Following Sec. 11.3 of "Numerical Recipies in C", W.H. Press, S.A.
   Teukolsky, W.T. Vetterling, B.P. Flannery, Cambridge University Press (1992)

   The diagonal matrix elements of the tri-diagonal matrix
   are now in d[0..n-1], while the vector e[0..n-2]
   contains the subdiagonal matrix elements of the tridiagonal matrix */

#ifdef DEBUGDIAG
   debug_out("tridiag_ql_implicit(), start of QL algorithm");
#endif
   
   for(l=0; l<dim; l++)
    {
     iter=0;
     do
      {
       for(m=l; m < dim-1; m++)		/* Look for a single small subdiagonal
					   element to split the matrix */
        {
         dd=fabs(d[m])+fabs(d[m+1]);
         if((fabs(e[m])+dd) == dd)
               break;
        }
       if(m != l)
        {
          if(iter++ >= 40 /* 30 */)
           {
            display_error("Diagonalization of tridiagonal matrix did not converge");
            goto end_it;
           }
          g=(d[l+1]-d[l])/(2.0*e[l]);	/* Form shift */
          /* replace this by non-destructive variant
          r = sqrt((g*g) + ((double) 1));
          */
          if((p = fabs(g)) > ((double) 1))
            r = p*sqrt(((double) 1)/(g*g) + ((double) 1));
          else
            r = sqrt((g*g) + ((double) 1));
          /* ... replacement ends here ... */
          if(g > 0)			/* Compute d_m - k_S */
            g = d[m]-d[l]+e[l]/(g+r);
          else
            g = d[m]-d[l]+e[l]/(g-r);
          s = c = 1.0L;
          p = 0;
          for(i=m-1; i>=l; i--)		/* A plane rotation as in the original
					   QL, followed by a Givens rotation
					   to restore tridiagonal form */
           {
            f = s*e[i];
            b = c*e[i];
            if(fabs(f) >= fabs(g))
             {
              c = g/f;
              r = sqrt((c*c) + ((double) 1));
              if(i < dim-2)
                e[i+1] = f*r;
              c *= (s = 1.0L/r);
             }
            else
             {
              s = f/g;
              r = sqrt((s*s) + ((double) 1));
              if(i < dim-2)
                e[i+1] = g*r;
              s *= (c = 1.0/r);
             }
            g = d[i+1]-p;
            r = (d[i]-g)*s+2.0L*c*b;
            p = s*r;
            d[i+1] = g+p;
            g=c*r-b;
            if(bchg)				/* suppress if not wanted */
             {
              /* It is better to apply the transformation later.
                 This allows the us to optimize cash performance and to
                 parallelize the necessary loop(s) */
              if(tqli_cnt >= TQLIBUFSIZE)
                tqli_buf_flush(bchg, dim);
              tqli_buf[tqli_cnt].i=i;
              tqli_buf[tqli_cnt].c=c;
              tqli_buf[tqli_cnt].s=s;
              tqli_cnt++;
             }
           }
          d[l] -= p;
          e[l] = g;
          if(m < dim-1)
            e[m] = 0;
        }
      } while (m != l);
    }

end_it:
   /* Bubble sort for eigenvalues */
#ifdef DEBUGDIAG
   debug_out("tridiag_ql_implicit(), start of bubble sort");
#endif

   if(bchg)
    {
     tqli_buf_flush(bchg, dim);			/* first finish transformations */
     free(tqli_buf);

     smap = (int *)malloc(dim * sizeof(int));
     if(! smap)
      {
        display_error("Not enough memory to sort eigenvalues");
        return;
      }
     for(i=0; i<dim; i++)
        smap[i] = i;
    }

   flag = TRUE;
   while(flag)
    {
      flag = FALSE;
      for(l=0; l<dim-1; l++)
       if(d[l] > d[l+1])
        {
          f = d[l];
          d[l] = d[l+1];
          d[l+1] = f;
	  if(bchg)
           {
	    /* We will shuffle columns later */
	    temp_int = smap[l];
            smap[l] = smap[l+1];
            smap[l+1] = temp_int;
/* *** old version
            swap_columns(bchg, l, l+1);
 */
           }
          flag = TRUE;
        }
    }

   if(bchg)
    {
     /* Now shuffle columns around */
     for(i=0; i<dim; i++)
      if((smap[i] >= 0) && (smap[i] != i))
       {
#pragma omp parallel for private(k,mnum,mnum1,re,im,bptr) if(dim>500)
          for(k=0; k<dim; k++)
  	   {
            /* Move out as many pointer computations as possible
               from inner loop */
	    bptr = bchg->entries[k];
	    re = bptr[i].re;
	    im = bptr[i].im;
            mnum = i;
            while(smap[mnum] != i)
             {
              mnum1 = smap[mnum];
	      bptr[mnum].re = bptr[mnum1].re;
	      bptr[mnum].im = bptr[mnum1].im;
              mnum = mnum1;
             }
	    bptr[mnum].re = re;
	    bptr[mnum].im = im;
	   }
	  /* mark this orbit as done */
          mnum = smap[i];
          while(mnum != i)
           {
            mnum1 = smap[mnum];
            smap[mnum] = -1;
            mnum = mnum1;
           }
       }
     free(smap);
    }

   if(res)
    {
     alloc_int_vector(res, dim);
     if(res->elements)
       for(l=0; l<dim; l++)
         res->elements[l].re = d[l];
    }

   free(e); 
   free(d);

#ifdef DEBUGDIAG
   debug_out("tridiag_ql_implicit() finished");
#endif
 }

#ifdef _OPENMP
extern int omp_numproc;
#endif

/* Use Householder method to reduce a Hermitian matrix to tridiagonl form
   inspired by Sec. 11.2 of "Numerical Recipes in C", W.H. Press, S.A.
   Teukolsky, W.T. Vetterling, B.P. Flannery, Cambridge University Press (1992)

   The input matrix is m. It is destroyed by this routine and returns
   the transformation matrix.
*/

#define HNAME "_householder_m"

static void householder(struct full_matrix *m, int bchg)
 {
  register int l, k, j, i;
  register REAL scale, hh, h;
  register struct matrix *tmat;
  struct vector pv;
  COMPLEX tmp, g, f, sigma;
  register REAL gre, gim;
  register REAL fre, fim;
  COMPLEX c1, c2;				/* a few complex numbers */
  register COMPLEX *cptr1, *cptr2, *cptr3;
#ifdef _OPENMP
  int process;
  struct vector hv[256];			/* Limit: 256 threads */
#endif
  
#ifdef DEBUGDIAG
  debug_out("householder() started");
#endif

  if((m->rows < 1) || (m->columns < 1))
   {
     display_error("Householder method called with improperly defined matrix");
     return;
   }

  if(m->rows != m->columns)
   {
     display_error("Householder method called with non-square matrix");
     return;
   }

  /* we need a temporary storage for the vector p */
  alloc_int_vector(&pv, m->rows);
#ifdef _OPENMP
  for(i=0; i<omp_numproc; i++)
    alloc_int_vector(hv+i, m->rows);		/* Each processor needs it private storage */
#endif

  /* allocate and identify result tridiagonal matrix */
  alloc_matrix((UBYTE *) HNAME, m->rows, m->columns);
  tmat = identify_matrix((UBYTE *) HNAME);

  for(i=m->rows-1; i>=1; i--)			/* start from the lower right corner of m */
   {
    l=i-1;
    h=scale=0;
    if(l > 0)
     {
      for(k=0; k<=l; k++)			/* sum row of m (eq. (11.2.19)) */
        scale += complex_abs(m->entries[i]+k);
      if(scale == 0)				/* skip transformation if this is zero */
       {
        /* copy matrix element below diagonal */
        insert_matrix_el(&(tmat->matrix_rows[i]), l, m->entries[i]+l);
        /* and complex conjugate above */ 
        complex_conjugate(m->entries[i]+l, &tmp);
        insert_matrix_el(&(tmat->matrix_rows[l]), i, &tmp);
       }
      else
       {
        for(k=0; k<=l; k++)
         {
          m->entries[i][k].re /= scale;		/* rescale entries of m (eq. (11.2.20)) */
          m->entries[i][k].im /= scale;
          /* and accumulate norm (squared) of "u" in h (to form sigma) */
          h += complex_abs_sq(m->entries[i]+k);
         }
        f.re = m->entries[i][l].re;		/* store entry m[i][i-1] in f */
        f.im = m->entries[i][l].im;

        /* we have to choose
           sigma = m[i][i-1]/m[i][i-1]^* |m[i]|^2 */
        complex_conjugate(&f, &c1);
        divide(&f, &c1, &sigma);
        sigma.re *= h;
        sigma.im *= h;

        /* g = +/- sqrt(sigma), where the sign is chosen
           such that |m[i][i-1] - g| is as large as possible */
        complex_sqrt(&sigma, &g);
        add(&f, &g, &c1);
        sub(&f, &g, &c2);
        if(complex_abs_sq(&c1) > complex_abs_sq(&c2))
         {
          g.re = -g.re;
          g.im = -g.im;
         }

        /* store scale*g below diagonal */
        c1.re = scale*g.re;
        c1.im = scale*g.im;
        insert_matrix_el(&(tmat->matrix_rows[i]), l, &c1);
        /* and scale*g^* above */
        complex_conjugate(&c1, &tmp);
        insert_matrix_el(&(tmat->matrix_rows[l]), i, &tmp);
        
        /* h = |m[i]|^2 - Re(g^* f) */
        h -= (g.re*f.re + g.im * f.im);		/* Now h is eq. (11.2.4) */
        /* substitute m[i][i-1] by 
           m[i][i-1] - g = f - g
           (=>  m[i] = u^t) */
        sub(&f, &g, m->entries[i]+l);

#ifdef DEBUGHOUSEHOLDER
        for(k=0; k<=l; k++)
         {
          printf("ut[%d][%d] = ", i+1, k+1);
          show_complex(m->entries[i]+k);
         }
        printf("h := %1.12lg;\n", h);
#endif

        fre = fim = 0;			/* f = 0 */

#pragma omp parallel for reduction(+:fre,fim) private(j,k,gre,gim,cptr1,cptr2) if(l>600)
        for(j=0; j<=l; j++)
         {
          /* The following two lines can be omitted if eigenvectors not wanted */
          if(bchg)
	   {
            m->entries[j][i].re = m->entries[i][j].re/h;		/* Store u^t/H in ith column of m */
            m->entries[j][i].im = m->entries[i][j].im/h;
           }

          /* Form an element of A*u in g */
          gre = gim = 0;			/* g = 0 */
          /* Move out as many pointer computations as possible
             from inner loop */
	  cptr1 = m->entries[i];
	  cptr2 = m->entries[j];
          for(k=0; k<=l; k++)
           {
            /* g += m[j][k]*m[i][k]^* */
/*
            multiply_c_add_re(m->entries[i]+k, m->entries[j]+k, gre);
            multiply_c_add_im(m->entries[i]+k, m->entries[j]+k, gim);
*/
            multiply_c_add_re(cptr1, cptr2, gre);
            multiply_c_add_im(cptr1, cptr2, gim);
	    cptr1++;
	    cptr2++;
           }
          pv.elements[j].re = gre/h;		/* Store the result in p[j] */
          pv.elements[j].im = gim/h;		/* (this is eq. (11.2.10) ) */
          /* f += m[i][j] * p[j] 
             - to accumulate u * p in f */
          multiply_add_re(m->entries[i]+j, pv.elements+j, fre);
          multiply_add_im(m->entries[i]+j, pv.elements+j, fim);
         }		/* end of   for(j=0; j<=l; j++) */
	f.re = fre;
	f.im = fim;
        if(! is_zero(fim/complex_abs(&f)) )
          display_error("Householder: K is not real, as expected");
        hh = fre/(h+h);				/* Form K, eq. (11.2.11) */

        for(j=0; j<=l; j++)
         {					/* Form q and overwrite p with it */
          /* store entry m[i][j]^* = u[j] in f */
          complex_conjugate(m->entries[i]+j, &f);
          /* p[j] = p[j] - hh*f */
          pv.elements[j].re -= hh*f.re;
          pv.elements[j].im -= hh*f.im;
	 }

#pragma omp parallel for private(j,k,f,g,c1,cptr1,cptr2,cptr3) if(l>600)
        for(j=0; j<=l; j++)
         {					/* Form q and overwrite p with it */
          /* store entry m[i][j]^* = u[j] in f */
          complex_conjugate(m->entries[i]+j, &f);
          /* g = p[j] */
          g.re = pv.elements[j].re;
          g.im = pv.elements[j].im;
          /* Reduce m, eq. (11.2.13) */
          /* Move out as many pointer computations as possible
             from inner loop */
	  cptr1 = pv.elements;
	  cptr2 = m->entries[j];
	  cptr3 = m->entries[i];
/* We could stop at j, but then we would have to get some matrix elements
   by complex conjugation later, which is not so good for cache performance */
          for(k=0; k<=l; k++)
           {
            /* m[j][k] -= (f * p[k]^* + g * m[i][k]) */
/*
            complex_conjugate(pv.elements+k, &c1);
            multiply_sub(&f, &c1, m->entries[j]+k);
            multiply_sub(&g, m->entries[i]+k, m->entries[j]+k);
*/
            complex_conjugate(cptr1, &c1);
            multiply_sub(&f, &c1, cptr2);
            multiply_sub(&g, cptr3, cptr2);
	    cptr1++;
	    cptr2++;
	    cptr3++;
           }
         }		/* end of   for(j=0; j<=l; j++) */
       }		/* end of   if(scale == 0) ... else */
     }			/* end of   if(l > 0) */
    else
     {
      /* copy matrix element below diagonal */
      insert_matrix_el(&(tmat->matrix_rows[i]), l, m->entries[i]+l);
      /* and complex conjugate above */
      complex_conjugate(m->entries[i]+l, &tmp);
      insert_matrix_el(&(tmat->matrix_rows[l]), i, &tmp);
     }
#ifdef DEBUGHOUSEHOLDER
    printf("Transformed matrix:\n");
    show_full_matrix(m);
#endif
    /* Now store h in p[i] - to be able to check later if it was zero */
    pv.elements[i].re = h;
    pv.elements[i].im = 0;
   }			/* end of   for(i=m->rows-1; i>=1; i--) */

  /* In the last transformation step, there is nothing to do */
  pv.elements[0].re = pv.elements[0].im = 0;

  /* undefined matrix elements in the result matrix are assumed to
     be zero, so there is no need to zero them */

#ifdef DEBUGDIAG
  debug_out("Started computing transformation in householder()");
#endif

  /* Contents of this loop can be omitted if eigenvectors not
     wanted except for statement d[i]=m[i][i]; */
  if(bchg)
   for(i=0; i<m->rows; i++)
   {
    l = i-1;
    if(pv.elements[i].re)
     {					/* skip block if we did not do a transformation */
#ifdef _OPENMP
#pragma omp parallel for private(process, j, k, cptr1, cptr2, cptr3) if(l>600)
      for(process=0; process<omp_numproc; process++)
       {
/* Swap j- and k- loops for better cache performance */
        for(j=0; j<=l; j++)
          hv[process].elements[j].re = hv[process].elements[j].im = 0;
        for(k=process; k<=l; k += omp_numproc)
         {
          /* Move out as many pointer computations as possible
             from inner loop */
          cptr1 = m->entries[i]+k;
	  cptr2 = m->entries[k];
	  cptr3 = hv[process].elements;
          for(j=0; j<=l; j++)
           {				/* Use u^t und u^t/H stored in m to form PQ */
            /* g += m[i][k]*m[k][j] */
/*            multiply_add(m->entries[i]+k, m->entries[k]+j, &(pv.elements[j])); */
             multiply_add(cptr1, cptr2, cptr3);
	     cptr2++;
	     cptr3++;
           }
         }
       }
/* collect results produced by different processes */
      for(j=0; j<=l; j++)
       {
        pv.elements[j].re = pv.elements[j].im = 0;
        for(process=0; process<omp_numproc; process++)
	 {
	    pv.elements[j].re += hv[process].elements[j].re;
	    pv.elements[j].im += hv[process].elements[j].im;
	 }
       }
#else
/* Swap j- and k- loops for better cache performance */
      for(j=0; j<=l; j++)
        pv.elements[j].re = pv.elements[j].im = 0;
      for(k=0; k<=l; k++)
       {
        /* Move out as many pointer computations as possible
           from inner loop */
        cptr1 = m->entries[i]+k;
	cptr2 = m->entries[k];
	cptr3 = pv.elements;
        for(j=0; j<=l; j++)
         {				/* Use u^t und u^t/H stored in m to form PQ */
          /* g += m[i][k]*m[k][j] */
/*          multiply_add(m->entries[i]+k, m->entries[k]+j, &(pv.elements[j])); */
           multiply_add(cptr1, cptr2, cptr3);
	   cptr2++;
	   cptr3++;
         }
       }
#endif
/* Now we need a second loop */
#pragma omp parallel for private(j, k, c1, cptr1, cptr2) if(l>600)
      for(k=0; k<=l; k++)
       {
        complex_conjugate(m->entries[k]+i, &c1);
        /* Move out as many pointer computations as possible
           from inner loop */
	cptr1 = pv.elements;
	cptr2 = m->entries[k];
        for(j=0; j<=l; j++)
         {
          /* m[k][j] -= g * m[k][i]^* */
/*          multiply_sub(&(pv.elements[j]), &c1, m->entries[k]+j); */
           multiply_sub(cptr1, &c1, cptr2);
	   cptr1++;
	   cptr2++;
         }
       }
     }
    /* !!! this statement must be kept in any case !!! */
    insert_matrix_el(&(tmat->matrix_rows[i]), i, m->entries[i]+i);
    m->entries[i][i].re = 1;		/* Reset row and column of m to */
    m->entries[i][i].im = 0;		/* identity matrix for next iteration */
    for(j=0; j<=l; j++)
     {
      m->entries[i][j].re = m->entries[i][j].im = 0;
      m->entries[j][i].re = m->entries[j][i].im = 0;
     }
   }
  else
   for(i=0; i<m->rows; i++)
    /* !!! this statement must be kept in any case !!! */
    insert_matrix_el(&(tmat->matrix_rows[i]), i, m->entries[i]+i);

#ifdef _OPENMP
  for(i=0; i<omp_numproc; i++)
    free_vector(hv+i);
#endif

  /* now free the space occupied by the temporary vector p */
  free_vector(&pv);

#ifdef DEBUGDIAG
  debug_out("householder() finished");
#endif
 }

/* The interface for the Householder routine followed by
   TQLI
   m on input:  Hermitean matrix to be diagonalized
   m on output: Change of basis effecting diagonalization
*/

void householder_diagonalize(struct full_matrix *m, struct vector *res)
 {
  struct matrix *tmat;
  int i, j;

  householder(m, res?FALSE:TRUE);

  tmat = identify_matrix((UBYTE *) HNAME);
  if(res)
    tridiag_ql_implicit(tmat, 0, res);
  else
    tridiag_ql_implicit(tmat, m, 0);
 }
